	using System;
	using System.ComponentModel;
	using System.Windows.Forms;
	using System.Drawing;
	using System.Drawing.Text;
	using System.Diagnostics;
	using System.Drawing.Drawing2D;
	using System.Runtime.InteropServices;
	
	namespace MyTabControl.Graphics
	{

	/****************************************************************************************
	This class is used for double buffering	
	****************************************************************************************/
	public class GraphicsBuffer
	{
		/************************************************************************************
		Variables declaration
		************************************************************************************/
		protected int hDIB;
		protected int hCompatDC;
		protected int hOldBit;
		protected int m_nWidth;
		protected int m_nHeight;
		protected System.Drawing.Graphics oBuffer;


		/************************************************************************************
		Used to retrieve a double buffer graphics object
		************************************************************************************/
		public System.Drawing.Graphics RequestBuffer(System.Drawing.Graphics oDest, int nWidth, int nHeight) 
		{
			return CreateBuffer(oDest, nWidth, nHeight);
		}


		/************************************************************************************
		Paints the double buffer contents to the destination graphics object
		************************************************************************************/
		public void PaintBuffer(System.Drawing.Graphics oDest, int x, int y) 
		{
			int rop = 0xcc0020; // RasterOp.SOURCE.GetRop();
			int hDC = oDest.GetHdc().ToInt32();
			try 
			{
				APIMethods.BitBlt(hDC, x, y, m_nWidth, m_nHeight, hCompatDC, 0, 0, rop); 
			}
			finally 
			{
				oDest.ReleaseHdc(new IntPtr(hDC));
			}
		}


		/************************************************************************************
		Dispose
		************************************************************************************/
		public void Dispose() 
		{
			Cleanup();
		}


		/************************************************************************************
		Releases all allocated objects
		************************************************************************************/
		private void Cleanup() 
		{
			if( oBuffer!=null )
			{
				oBuffer.Dispose();
				oBuffer = null;
			}
			if( hOldBit!=(int)0 )
			{
				APIMethods.SelectObject(hCompatDC, hOldBit);
				hOldBit = (int)0;
			}
			if( hCompatDC!=(int)0 )
			{
				APIMethods.DeleteDC(hCompatDC);
				hCompatDC = (int)0;
			}
			if( hDIB!=(int)0 ) 
			{
				APIMethods.DeleteObject(hDIB);
				hDIB = (int)0;
			}
		}


		/************************************************************************************
		Creates a double buffer graphics object
		************************************************************************************/
		private System.Drawing.Graphics CreateBuffer(System.Drawing.Graphics oCompat, int nWidth, int nHeight) 
		{
			// Is the object already created?
			if( nWidth==m_nWidth && nHeight==m_nHeight && oBuffer!=null ) return oBuffer;

			Cleanup();

			int hSrcDC = oCompat.GetHdc().ToInt32();
			try 
			{
				int pvbits = APIMethods.Nullint;
				hDIB = CreateCompatibleDIB(hSrcDC, APIMethods.Nullint, nWidth, nHeight, ref pvbits);
				hCompatDC = APIMethods.CreateCompatibleDC(hSrcDC);
			}
			finally 
			{
				oCompat.ReleaseHdc(new IntPtr(hSrcDC));
			}

			hOldBit = APIMethods.SelectObject(hCompatDC, hDIB);
			oBuffer = System.Drawing.Graphics.FromHdc(new IntPtr(hCompatDC));
			m_nWidth = nWidth;
			m_nHeight = nHeight;

			return oBuffer;
		}


		/************************************************************************************
		Creates a DBI section compatible with the specified hDC
		************************************************************************************/
		private int CreateCompatibleDIB(int hDC, int hPal, int nWidth, int nHeight, ref int ppvBits) 
		{
			// Validate hdc.
			if( hDC==APIMethods.Nullint) throw new ArgumentNullException("hDC");
			if( APIMethods.GetObjectType(hDC)!=APIMethods.OBJ_DC ) throw new ArgumentException("hDC", "hDC must be screen DC!");

			int hDIBRet = APIMethods.Nullint;

			// Create a bitmap info structure
			APIMethods.BITMAPINFO_FLAT tBitInfo = new APIMethods.BITMAPINFO_FLAT();
			
			// Fill the bitmap info structure
			if( FillBitmapInfo(hDC, hPal, ref tBitInfo) ) 
			{
				// Change bitmap size to match specified dimensions.
				tBitInfo.bmiHeader_biWidth = nWidth;
				tBitInfo.bmiHeader_biHeight = nHeight;
				if (tBitInfo.bmiHeader_biCompression == APIMethods.BI_RGB) 
				{
					tBitInfo.bmiHeader_biSizeImage = 0;
				}
				else
				{
					if ( tBitInfo.bmiHeader_biBitCount == 16 )			tBitInfo.bmiHeader_biSizeImage = nWidth * nHeight * 2;
					else if ( tBitInfo.bmiHeader_biBitCount == 32 )		tBitInfo.bmiHeader_biSizeImage = nWidth * nHeight * 4;
					else												tBitInfo.bmiHeader_biSizeImage = 0;
				}
				tBitInfo.bmiHeader_biClrUsed = 0;
				tBitInfo.bmiHeader_biClrImportant = 0;

				// Create the DIB section.  Let Win32 allocate the memory and return
				// a pointer to the bitmap surface.
				hDIBRet = APIMethods.CreateDIBSection(hDC, ref tBitInfo, APIMethods.DIB_RGB_COLORS, ref ppvBits, APIMethods.Nullint, 0);
				if( hDIBRet==APIMethods.Nullint ) throw new Win32Exception(Marshal.GetLastWin32Error());
			}

			return hDIBRet;

		}


		/************************************************************************************
		Fills a bitmap information structure
		************************************************************************************/
		private bool FillBitmapInfo(int hDC, int hPal, ref APIMethods.BITMAPINFO_FLAT tBitInfo)
		{
			int hBit = APIMethods.Nullint;
			bool bRet = false;

			try 
			{

				// Create a dummy bitmap from which we can query color format info
				// about the device surface.
				hBit = APIMethods.CreateCompatibleBitmap(hDC, 1, 1);
				if( hBit==APIMethods.Nullint ) throw new Exception("Failed to create bitmap for query");
	
				// Structure size and number of colors
				tBitInfo.bmiHeader_biSize = Marshal.SizeOf(typeof(APIMethods.BITMAPINFOHEADER));
				tBitInfo.bmiColors = new byte[APIMethods.BITMAPINFO_MAX_COLORSIZE*4];

				// Call first time to fill in BITMAPINFO header.
				APIMethods.GetDIBits(hDC, hBit, 0, 0, APIMethods.Nullint, ref tBitInfo, APIMethods.DIB_RGB_COLORS);
				
				// Number of bits
				if( tBitInfo.bmiHeader_biBitCount <= 8 ) 
				{
					bRet = FillColorTable(hDC, hPal, ref tBitInfo);
				}
				else 
				{
					if ( tBitInfo.bmiHeader_biCompression==APIMethods.BI_BITFIELDS ) 
					{
						// Call a second time to get the color masks.
						// It's a GetDIBits Win32 "feature".
						APIMethods.GetDIBits(hDC, hBit, 0, tBitInfo.bmiHeader_biHeight, APIMethods.Nullint, ref tBitInfo, APIMethods.DIB_RGB_COLORS);
					}
					bRet = true;
				}
			}
			finally
			{
				// Release the dummy bitmap
				if( hBit!=APIMethods.Nullint )
				{
					APIMethods.DeleteObject(hBit);
					hBit = APIMethods.Nullint;
				}
			}

			return bRet;
		}


		/************************************************************************************
		Fills a bitmap color table
		************************************************************************************/
		private unsafe bool FillColorTable(int hDC, int hPal, ref APIMethods.BITMAPINFO_FLAT tBitInfo) 
		{
			bool bRet = false;
			byte[] aj = new byte[sizeof(APIMethods.PALETTEENTRY) * 256];
			int nColors;
			/*try
			{
				fixed (APIMethods.RGBQUAD* prgb = (APIMethods.RGBQUAD*) &tBitInfo.bmiColors)
				{
					fixed (APIMethods.PALETTEENTRY* lppe = (APIMethods.PALETTEENTRY*) &aj)
					{
						nColors = 1 << tBitInfo.bmiHeader_biBitCount;
						if( nColors <= 256 ) 
						{
							int hPalRet;
							if( hPal==APIMethods.Nullint )	hPalRet = APIMethods.GetSystemPaletteEntries(hDC, 0, nColors, aj);
							else							hPalRet = APIMethods.GetPaletteEntries(hPal, 0, nColors, aj);

							if( hPalRet!=0 ) 
							{
								for( int nLoop=0; nLoop<nColors; nLoop++ )
								{
									prgb[nLoop].rgbRed      = lppe[nLoop].peRed;
									prgb[nLoop].rgbGreen    = lppe[nLoop].peGreen;
									prgb[nLoop].rgbBlue     = lppe[nLoop].peBlue;
									prgb[nLoop].rgbReserved = 0;
								}

								bRet = true;
							}
						}
					}

				}
			}
			catch(Exception exp)
			{
				MessageBox.Show(exp.ToString());
			}*/
			return bRet;
		}
	}


	/****************************************************************************************
	Helper class used to access the windows API
	****************************************************************************************/
	class APIMethods 
	{		
		/************************************************************************************
		Constants
		************************************************************************************/
		public static readonly int Nullint = (int)0;
		public const int        BITMAPINFO_MAX_COLORSIZE = 256,
			DIB_RGB_COLORS = 0,
			DIB_PAL_COLORS = 1,
			OBJ_PEN = 1,
			OBJ_BRUSH = 2,
			OBJ_DC = 3,
			OBJ_METADC = 4,
			OBJ_PAL = 5,
			OBJ_FONT = 6,
			OBJ_BITMAP = 7,
			OBJ_REGION = 8,
			OBJ_METAFILE = 9,
			OBJ_MEMDC = 10,
			OBJ_EXTPEN = 11,
			OBJ_ENHMETADC = 12,
			OBJ_ENHMETAFILE = 13,
			BI_RGB = 0,
			BI_RLE8 = 1,
			BI_RLE4 = 2,
			BI_BITFIELDS = 3,
			PLANES = 14,
			BITSPIXEL = 12;

		/************************************************************************************
		Structures
		************************************************************************************/
		public struct RGBQUAD 
		{
			public byte		rgbBlue;
			public byte		rgbGreen;
			public byte		rgbRed;
			public byte		rgbReserved;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class BITMAPINFOHEADER 
		{
			public int      biSize = Marshal.SizeOf(typeof(BITMAPINFOHEADER));
			public int      biWidth;
			public int      biHeight;
			public short    biPlanes;
			public short    biBitCount;
			public int      biCompression;
			public int      biSizeImage;
			public int      biXPelsPerMeter;
			public int      biYPelsPerMeter;
			public int      biClrUsed;
			public int      biClrImportant;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class BITMAPINFO 
		{
			public BITMAPINFOHEADER bmiHeader = new BITMAPINFOHEADER();

			[MarshalAs(System.Runtime.InteropServices.UnmanagedType.ByValArray, SizeConst=BITMAPINFO_MAX_COLORSIZE*4)]
			public byte[] bmiColors; // RGBQUAD structs... Blue-Green-Red-Reserved, repeat...
		}

		[StructLayout(LayoutKind.Sequential)]
			public struct PALETTEENTRY 
		{
			public byte		peRed;
			public byte		peGreen;
			public byte		peBlue;
			public byte		peFlags;
		}

		[StructLayout(LayoutKind.Sequential)]
			public struct BITMAPINFO_FLAT 
		{
			public int      bmiHeader_biSize;
			public int      bmiHeader_biWidth;
			public int      bmiHeader_biHeight;
			public short    bmiHeader_biPlanes;
			public short    bmiHeader_biBitCount;
			public int      bmiHeader_biCompression;
			public int      bmiHeader_biSizeImage;
			public int      bmiHeader_biXPelsPerMeter;
			public int      bmiHeader_biYPelsPerMeter;
			public int      bmiHeader_biClrUsed;
			public int      bmiHeader_biClrImportant;

			[MarshalAs(System.Runtime.InteropServices.UnmanagedType.ByValArray, SizeConst=BITMAPINFO_MAX_COLORSIZE*4)]
			public byte[] bmiColors; // RGBQUAD structs... Blue-Green-Red-Reserved, repeat...
		}

		[DllImport("gdi32")]
		public static extern int GetPaletteEntries(int hpal, int iStartIndex, int nEntries, byte[] lppe);
		[DllImport("gdi32")]
		public static extern int GetSystemPaletteEntries(int hdc, int iStartIndex, int nEntries, byte[] lppe);
		[DllImport("gdi32")]
		public static extern int CreateDIBSection(int hdc, ref APIMethods.BITMAPINFO_FLAT bmi, int iUsage, ref int ppvBits, int hSection, int dwOffset);
		[DllImport("gdi32")]
		public static extern int GetObjectType(int hobject);
		[DllImport("gdi32")]
		public static extern int CreateCompatibleDC(int hDC);
		[DllImport("gdi32")]
		public static extern int CreateCompatibleBitmap(int hDC, int width, int height);
		[DllImport("gdi32")]
		public static extern int GetDIBits(int hdc, int hbm, int arg1, int arg2, int arg3, APIMethods.BITMAPINFOHEADER bmi, int arg5);
		[DllImport("gdi32")]
		public static extern int GetDIBits(int hdc, int hbm, int arg1, int arg2, int arg3, ref APIMethods.BITMAPINFO_FLAT bmi, int arg5);
		[DllImport("gdi32")]
		public static extern int SelectObject(int hdc, int obj);
		[DllImport("gdi32")]
		public static extern bool DeleteObject(int hObject);
		[DllImport("gdi32")]
		public static extern bool DeleteDC(int hDC);
		[DllImport("gdi32")]
		public static extern bool BitBlt(int hDC, int x, int y, int nWidth, int nHeight, int hSrcDC, int xSrc, int ySrc, int dwRop);
		[DllImport("gdi32")]
		public static extern int GetDeviceCaps(int hDC, int nIndex);
	}

}