using System;
using System.Data;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections;
using System.ComponentModel;
using System.ComponentModel.Design;	

namespace MyTabControl
{


	/*****************************************************************************************
	Enumerators
	*****************************************************************************************/
	public enum TabBorders { LeftOuter=2, LeftInner=4, TopOuter=8, TopInner=16, RightOuter=32, RightInner=64, BottomOuter=128, BottomInner=256, All=510, None=0 }
	public enum TabState { Normal=0, Active=2, Hot=4, Invisible=8 }
	public enum TabStyle { None = 0, VisualStudio = 1, InterDev = 2, Excel = 3, Photoshop = 4 }



	/*****************************************************************************************
	Tab object
	*****************************************************************************************/
	//[ShowInToolbox(false)]
	public class Tab : Component//System.Windows.Forms.TabPage
	{
		/*************************************************************************************
		Variables Declaration
		*************************************************************************************/

		// Property variables
		protected string		m_sText				= "Tab";
		protected int			m_nX				= 0,
								m_nY				= 0,
								m_nWidth			= 40,
								m_nHeight			= 0,
								m_nImage			= -1,
								m_nHotImage			= -1;
		protected short			m_nSlope			= 0,
								m_nSlopePhoto		= 0,
								m_nBorderSpace		= 3,
								m_nImgSpace			= 0;
		protected bool			m_bAutoWidth		= true,
								m_bOverlaySlope		= false,
								m_bVisible			= true;
		protected Color			m_LOColor			= Color.Transparent,
								m_LIColor			= Color.Transparent,
								m_TOColor			= Color.Transparent,
								m_TIColor			= Color.Transparent,
								m_ROColor			= Color.DarkGray,
								m_RIColor			= Color.Transparent,
								m_BOColor			= Color.Transparent,
								m_BIColor			= Color.Transparent,
								m_ALOColor			= Color.White,
								m_ALIColor			= Color.Transparent,
								m_ATOColor			= Color.Transparent,
								m_ATIColor			= Color.Transparent,
								m_AROColor			= Color.Black,
								m_ARIColor			= Color.Transparent,
								m_ABOColor			= Color.Black,
								m_ABIColor			= Color.FromArgb(247, 243, 233),
								m_BackColor			= Color.FromArgb(247, 243, 233) /*SystemColors.Control*/,
								m_HotBackColor		= Color.FromArgb(247, 243, 233),
								m_ActiveBackColor	= SystemColors.Control,
								m_ForeColor			= Color.DarkGray,
								m_HotForeColor		= Color.Blue,
								m_ActiveForeColor	= SystemColors.ControlText;
		protected Font			m_Font				= null,
								m_HotFont			= null,
								m_ActiveFont		= null;
		protected StringFormat	m_TextFormat		= null;
		protected TabBorders	m_nBorderSides		= TabBorders.None,
								m_nABorderSides		= TabBorders.All;
		protected TabStyle		m_nStyle			= TabStyle.None;
		protected ContentAlignment m_nImgAlign		= ContentAlignment.MiddleLeft;

		// Internal variables
		internal TabCtl			m_oParent			= null;						// Tab parent control
		internal int			m_nIndex			= -1;						// Tab index id
		protected int			m_nRealWidth		= -1;						// The real width, seful when the control is in auto size mode
		protected bool			m_bSizeChanged		= true;						// True if the tab size has changed since the last update
		protected Region		m_oRgn				= null;						// Tab region, used to drawing and hittest
		protected Point[]		m_aPoints			= null;						// Tab coordinates
		protected TabState		m_nState			= TabState.Invisible;		// The current tab state
		protected Rectangle		m_ImgArea,										// Image area
								m_TxArea;										// Text area
		protected int			m_nMinWidth			= -1;						// The minimun width allowed
		protected int			m_nMaxWidth			= -1;						// The max width allow by the tab control
		protected TabDirection	m_nDirection		= TabDirection.Top;			// Direction of tab

		/*************************************************************************************
		Constructor
		*************************************************************************************/
		public Tab()
		{
			// Initialize default objects and properties
			m_TextFormat = new StringFormat(StringFormatFlags.NoWrap);
			m_TextFormat.Alignment = StringAlignment.Center;
			m_TextFormat.LineAlignment = StringAlignment.Center;
		}

		public Tab(string sText) : this()
		{
			m_sText = sText;
		}

		public Tab(string sText, Font oFont) : this()
		{
			m_Font = oFont;
			m_sText = sText;
		}


		/*************************************************************************************
		Finalize
		*************************************************************************************/
		protected void Finalize()
		{
			if( m_ActiveFont != null ) m_ActiveFont.Dispose();
			if( m_HotFont != null ) m_HotFont.Dispose();
			if( m_Font != null ) m_Font.Dispose();
			if( m_oRgn != null ) m_oRgn.Dispose();
			if( m_TextFormat != null ) m_TextFormat.Dispose();		
			m_ActiveFont = null;
			m_HotFont = null;
			m_Font = null;
			m_oRgn = null;
			m_TextFormat = null;
			m_aPoints = null;

			//base.Finalize();
		}


		/*************************************************************************************
		Index property
		*************************************************************************************/
		[Category("Behavior"), Browsable(false)]//, Persistable(PersistableSupport.None)]
		public virtual int Index 
		{
			get { return m_nIndex; }
			set { if( m_oParent!=null ) m_oParent.Tabs.SetObjectIndex(this, value); }
		}

		/*************************************************************************************
		Style property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(TabStyle.None)]
		public virtual TabStyle Style 
		{
			get { return m_nStyle; }
			set 
			{ 
				if( m_nStyle == value ) return;
				m_nStyle = value; 
				UpdateStyle();
				PropertyChange("TabStyle", true);
			}
		}	

		/*************************************************************************************
		Visible property
		*************************************************************************************/
		[Category("Behavior"), Bindable(false), DefaultValue(true)]
		public virtual bool Visible 
		{
			get { return m_bVisible; }
			set 
			{ 
				m_bVisible = value; 
				PropertyChange("Visible", true);
			}
		}

		/*************************************************************************************
		Text property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue("Tab")]
		public virtual string Text 
		{
			get { return m_sText; }
			set 
			{ 
				m_sText = value; 
				PropertyChange("Text", true);
			}
		}

		/*************************************************************************************
		ImageIndex property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(-1)]
		public virtual int ImageIndex 
		{
			get { return m_nImage; }
			set 
			{ 
				m_nImage = value; 
				PropertyChange("ImageIndex", true);
			}
		}

		/*************************************************************************************
		HotImageIndex property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(-1)]
		public virtual int HotImageIndex 
		{
			get { return m_nHotImage; }
			set 
			{ 
				m_nHotImage = value; 
				PropertyChange("HotImageIndex", true);
			}
		}

		/*************************************************************************************
		ImageSpace property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(0)]
		public virtual int ImageSpace 
		{
			get { return m_nImgSpace; }
			set 
			{
				m_nImgSpace = (short)value;
				PropertyChange("ImageSpace", true);
			}
		}

		/*************************************************************************************
		ImageAlign property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(ContentAlignment.MiddleLeft)]
		public virtual ContentAlignment ImageAlign 
		{
			get { return m_nImgAlign; }
			set 
			{
				m_nImgAlign = value;
				PropertyChange("ImageAlign", true);
			}
		}

		/*************************************************************************************
		Width property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual int Width 
		{
			get 
			{
				if( m_bSizeChanged ) UpdateAreas(-1);
				return m_nWidth; 
			}
			set 
			{ 
				if( m_bAutoWidth==true ) return;
				if( value<=0 ) throw(new ArgumentException("Width must be greater then 0"));
				m_nWidth = value; 
				PropertyChange("Width", true);
			}
		}
		public bool ShouldPersistWidth() { return (m_bAutoWidth==false && m_nWidth!=40 ); }
		public void ResetWidth() { m_nWidth=40; m_bSizeChanged=true; }

		// Actual width of the tab
		[Browsable(false)]
		public virtual int RealWidth 
		{ 
			get 
			{ 
				if( m_bSizeChanged ) UpdateAreas(-1);
				return m_nRealWidth; 
			}	
		}
		// Min width
		[Browsable(false)]
		public virtual int MinWidth { get { return m_nMinWidth; }	}
		// Max width
		[Browsable(false)]
		public virtual int MaxWidth { get { return m_nMaxWidth; }	}

		/*************************************************************************************
		Slope property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual int Slope 
		{
			get { return m_nSlope; }
			set 
			{ 
				m_nSlope = Int16.Parse(value.ToString()); 
				PropertyChange("Slope", true);
			}
		}
		public bool ShouldPersistSlope() 
		{ 
			if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev ) return m_nSlope!=15;
			return m_nSlope!=0;
		}
		public void ResetSlope() 
		{ 
			if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev ) Slope = 15;
			else Slope=0;
		}

		/*************************************************************************************
		AutoWidth property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(true)]
		public virtual bool AutoWidth 
		{
			get { return m_bAutoWidth; }
			set 
			{ 
				m_bAutoWidth = value; 
				if( m_bAutoWidth )		m_nWidth = 0;
				else if( m_nWidth==0 )	m_nWidth = 25;
				PropertyChange("AutoWidth", true);
			}
		}

		/*************************************************************************************
		BackColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color BackColor 
		{
			get { return m_BackColor; }
			set 
			{ 
				m_BackColor = value; 
				PropertyChange("BackColor", false);
			}
		}
		public bool ShouldPersistBackColor() 
		{ 
			if( m_nStyle==TabStyle.Excel ) return m_BackColor!=SystemColors.Control;
			if(  m_nStyle==TabStyle.InterDev ) return m_BackColor!=Color.DarkGray;
			return m_BackColor!=Color.FromArgb(247, 243, 233);
		}
		public void ResetBackColor() 
		{
			if( m_nStyle==TabStyle.Excel ) BackColor = SystemColors.Control;
			else if(  m_nStyle==TabStyle.InterDev ) BackColor = Color.DarkGray;
			else BackColor = Color.FromArgb(247, 243, 233);
		}

		/*************************************************************************************
		HotBackColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color HotBackColor 
		{
			get { return m_HotBackColor; }
			set 
			{ 
				m_HotBackColor = value; 
				PropertyChange("HotBackColor", false);
			}
		}
		public bool ShouldPersistHotBackColor() 
		{ 
			if( m_nStyle==TabStyle.Excel ) return m_HotBackColor!=SystemColors.Control;
			if(  m_nStyle==TabStyle.InterDev ) return m_HotBackColor!=Color.DarkGray;
			return m_HotBackColor!=Color.FromArgb(247, 243, 233);
		}
		public void ResetHotBackColor() 
		{ 
			if( m_nStyle==TabStyle.Excel ) HotBackColor = SystemColors.Control;
			else if(  m_nStyle==TabStyle.InterDev ) HotBackColor = Color.DarkGray;
			else HotBackColor = Color.FromArgb(247, 243, 233);
		}

		/*************************************************************************************
		ActiveBackColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color ActiveBackColor 
		{
			get { return m_ActiveBackColor; }
			set 
			{ 
				m_ActiveBackColor = value; 
				PropertyChange("ActiveBackColor", false);
			}
		}
		public bool ShouldPersistActiveBackColor() 
		{ 
			if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev ) return m_ActiveBackColor!=Color.White;
			return m_ActiveBackColor!=SystemColors.Control;
		}
		public void ResetActiveBackColor() 
		{
			if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev ) ActiveBackColor = Color.White;
			else ActiveBackColor = SystemColors.Control;
		}

		/*************************************************************************************
		ForeColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color ForeColor 
		{
			get { return m_ForeColor; }
			set 
			{ 
				m_ForeColor = value; 
				PropertyChange("ForeColor", false);
			}
		}
		public bool ShouldPersistForeColor() 
		{ 
			if( m_nStyle==TabStyle.InterDev ) return m_ForeColor!=Color.White;
			if( m_nStyle==TabStyle.Excel ) return m_ForeColor!=SystemColors.ControlText;
			return m_ForeColor!=Color.DarkGray;
		}
		public void ResetForeColor() 
		{ 
			if( m_nStyle==TabStyle.InterDev ) ForeColor = Color.White;
			if( m_nStyle==TabStyle.Excel ) ForeColor = SystemColors.ControlText;
			else ForeColor = Color.DarkGray;
		}

		/*************************************************************************************
		HotForeColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color HotForeColor 
		{
			get { return m_HotForeColor; }
			set 
			{ 
				m_HotForeColor = value; 
				PropertyChange("HotForeColor", false);
			}
		}
		public bool ShouldPersistHotForeColor() 
		{ 
			if( m_nStyle==TabStyle.InterDev ) return m_HotForeColor!=Color.Black;
			return m_HotForeColor!=Color.Blue;
		}
		public void ResetHotForeColor() 
		{ 
			if( m_nStyle==TabStyle.InterDev ) HotForeColor = Color.Black;
			else HotForeColor = Color.Blue;
		}

		/*************************************************************************************
		ActiveForeColor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual Color ActiveForeColor 
		{
			get { return m_ActiveForeColor; }
			set 
			{ 
				m_ActiveForeColor = value; 
				PropertyChange("ActiveForeColor", false);
			}
		}
		public bool ShouldPersistActiveForeColor() { return !(m_ActiveForeColor==SystemColors.ControlText); }
		public void ResetActiveForeColor() { ActiveForeColor=SystemColors.ControlText; }

		/*************************************************************************************
		Font property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual System.Drawing.Font Font 
		{
			get { return m_Font; }
			set 
			{ 
				if( m_Font!=null ) m_Font.Dispose();
				m_Font = value;
				PropertyChange("Font", true);
			}
		}
		public bool ShouldPersistFont() { return m_Font!=null; }
		public void ResetFont() { Font = null; }

		/*************************************************************************************
		HotFont property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual System.Drawing.Font HotFont 
		{
			get { return m_HotFont; }
			set 
			{ 
				if( m_HotFont!=null ) m_HotFont.Dispose();
				m_HotFont = value; 
				PropertyChange("HotFont", false);
			}
		}
		public bool ShouldPersistHotFont() { return m_HotFont!=null; }
		public void ResetHotFont() { HotFont = null; }

		/*************************************************************************************
		ActiveFont property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual System.Drawing.Font ActiveFont 
		{
			get { return m_ActiveFont; }
			set 
			{ 
				if( m_ActiveFont!=null ) m_ActiveFont.Dispose();
				m_ActiveFont = value; 
				PropertyChange("ActiveFont", false);
			}
		}
		public bool ShouldPersistActiveFont() { return m_ActiveFont!=null; }
		public void ResetActiveFont() { ActiveFont = null; }

		/*************************************************************************************
		BorderSides property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false)]
		public virtual TabBorders BorderSides 
		{
			get { return m_nBorderSides; }
			set 
			{
				m_nBorderSides = value; 
				PropertyChange("BorderSides", false);
			}
		}
		public bool ShouldPersistBorderSides() 
		{ 
			if( m_nStyle==TabStyle.InterDev || m_nStyle==TabStyle.Excel ) return m_nBorderSides!=TabBorders.All;
			return m_nBorderSides!=TabBorders.None;
		}
		public void ResetBorderSides() 
		{ 
			if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev ) BorderSides=TabBorders.All;
			else BorderSides=TabBorders.None;
		}

		/*************************************************************************************
		ActiveBorderSides property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(TabBorders.All)]
		public virtual TabBorders ActiveBorderSides 
		{
			get { return m_nABorderSides; }
			set 
			{
				m_nABorderSides = value; 
				PropertyChange("ActiveBorderSides", false);
			}
		}

		/*************************************************************************************
		OverlaySlope property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(false)]
		public virtual bool OverlaySlope 
		{
			get { return m_bOverlaySlope; }
			set 
			{ 
				m_bOverlaySlope = value; 
				PropertyChange("OverlaySlope", true);
			}
		}

		/*************************************************************************************
		BorderSpace property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(3)]
		public virtual int BorderSpace 
		{
			get { return m_nBorderSpace; }
			set 
			{ 
				m_nBorderSpace = (short)value; 
				PropertyChange("BorderSpace", true);
			}
		}

		/*************************************************************************************
		TextFormat property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(null)]
		public virtual StringFormat TextFormat 
		{
			get { return m_TextFormat; }
			set 
			{ 
				if( m_TextFormat!=null ) m_TextFormat.Dispose();
				m_TextFormat = value; 
				PropertyChange("TextFormat", false);
			}
		}

		/*************************************************************************************
		Set/Get border colors
		*************************************************************************************/
		public virtual void BorderColor(TabBorders nBorder, Color nColor)
		{
			switch (nBorder)
			{
				case TabBorders.LeftOuter:		m_LOColor = nColor;		break;
				case TabBorders.LeftInner:		m_LIColor = nColor;		break;
				case TabBorders.TopOuter:		m_TOColor = nColor;		break;
				case TabBorders.TopInner:		m_TIColor = nColor;		break;
				case TabBorders.RightOuter:		m_ROColor = nColor;		break;
				case TabBorders.RightInner:		m_RIColor = nColor;		break;
				case TabBorders.BottomOuter:	m_BOColor = nColor;		break;
				case TabBorders.BottomInner:	m_BIColor = nColor;		break;
				default:					throw(new ArgumentException("Invalid tab border"));
			}
		}
		public virtual Color BorderColor(TabBorders nBorder)
		{
			Color nReturn;

			switch (nBorder)
			{
				case TabBorders.LeftOuter:		nReturn = m_LOColor;	break;
				case TabBorders.LeftInner:		nReturn = m_LIColor;	break;
				case TabBorders.TopOuter:		nReturn = m_TOColor;	break;
				case TabBorders.TopInner:		nReturn = m_TIColor;	break;
				case TabBorders.RightOuter:		nReturn = m_ROColor;	break;
				case TabBorders.RightInner:		nReturn = m_RIColor;	break;
				case TabBorders.BottomOuter:	nReturn = m_BOColor;	break;
				case TabBorders.BottomInner:	nReturn = m_BIColor;	break;
				default:					throw(new ArgumentException("Invalid tab border"));
			}
	
			PropertyChange("BorderColor", false);
			return nReturn;
		}

		/*************************************************************************************
		Set/Get a active border color
		*************************************************************************************/
		public virtual void ActiveBorderColor(TabBorders nBorder, Color nColor)
		{
			switch (nBorder)
			{
				case TabBorders.LeftOuter:		m_ALOColor = nColor;	break;
				case TabBorders.LeftInner:		m_ALIColor = nColor;	break;
				case TabBorders.TopOuter:		m_ATOColor = nColor;	break;
				case TabBorders.TopInner:		m_ATIColor = nColor;	break;
				case TabBorders.RightOuter:		m_AROColor = nColor;	break;
				case TabBorders.RightInner:		m_ARIColor = nColor;	break;
				case TabBorders.BottomOuter:	m_ABOColor = nColor;	break;
				case TabBorders.BottomInner:	m_ABIColor = nColor;	break;
				default:					throw(new ArgumentException("Invalid tab border"));
			}
		}
		public virtual Color ActiveBorderColor(TabBorders nBorder)
		{
			Color nReturn;

			switch (nBorder)
			{
				case TabBorders.LeftOuter:		nReturn = m_ALOColor;	break;
				case TabBorders.LeftInner:		nReturn = m_ALIColor;	break;
				case TabBorders.TopOuter:		nReturn = m_ATOColor;	break;
				case TabBorders.TopInner:		nReturn = m_ATIColor;	break;
				case TabBorders.RightOuter:		nReturn = m_AROColor;	break;
				case TabBorders.RightInner:		nReturn = m_ARIColor;	break;
				case TabBorders.BottomOuter:	nReturn = m_ABOColor;	break;
				case TabBorders.BottomInner:	nReturn = m_ABIColor;	break;
				default:					throw(new ArgumentException("Invalid tab border"));
			}
	
			PropertyChange("ActiveBorderColor", false);
			return nReturn;
		}

		/*************************************************************************************
		Returns the tab state
		*************************************************************************************/
		[Browsable(false)]
		public virtual TabState State 
		{
			get { return m_nState; }
		}
		[Browsable(false)]
		public bool IsActive 
		{
			get { return (m_nState & TabState.Active) == TabState.Active; }
		}
		[Browsable(false)]
		public bool IsInvisible 
		{
			get { return (m_nState & TabState.Invisible) == TabState.Invisible; }
		}
		[Browsable(false)]
		public bool IsHot 
		{
			get { return (m_nState & TabState.Hot) == TabState.Hot; }
		}
		[Browsable(false)]
		public bool IsNormal 
		{
			get { return m_nState==TabState.Normal; }
		}

		/*************************************************************************************
		Returns the parent (tab control)
		*************************************************************************************/
		[Browsable(false)]
		public TabCtl Parent 
		{
			get { return m_oParent; }
			//set { m_oParent = value; }
		}

		/*************************************************************************************
		Returns the region
		*************************************************************************************/
		[Browsable(false)]
		public System.Drawing.Region Region 
		{
			get { return m_oRgn; }
		}

		/*************************************************************************************
		Returns the coordinates array
		*************************************************************************************/
		[Browsable(false)]
		public virtual Point[] Coordinates 
		{
			get { return m_aPoints; }		
		}
		[Browsable(false)]
		public virtual int X 
		{
			get { return m_nX; }		
		}
		[Browsable(false)]
		public virtual int Y 
		{
			get { return m_nY; }		
		}
		[Browsable(false)]
		public virtual int Height 
		{
			get { return m_nHeight; }		
		}

		/*************************************************************************************
		Returns the tab position
		*************************************************************************************/
		[Browsable(false)]
		public virtual Point Location
		{
			get { return new Point(m_nX, m_nY); }
		}

		/*************************************************************************************
		Returns the tab position
		*************************************************************************************/
		[Browsable(false)]
		public virtual Size Size
		{
			get { return new Size(m_nWidth, m_nHeight); }
		}


		/*************************************************************************************
		Returns true if the specified point is within the tab region
		*************************************************************************************/
		public virtual bool HitTest(int X, int Y)
		{
			if( m_oRgn==null ) return false;
			return m_oRgn.IsVisible(X, Y);
		}
		public virtual bool HitTest(Point pos)
		{
			if( m_oRgn==null ) return false;
			return m_oRgn.IsVisible(pos);
		}


		/*************************************************************************************
		Called when a property is changed
		*************************************************************************************/
		protected virtual void PropertyChange(string sProperty, bool bSizeChanged)
		{
			if( bSizeChanged ) m_bSizeChanged = true;

			// Informe the parent
			if( m_oParent!=null ) m_oParent.TabChanged(this, sProperty, bSizeChanged);
			return;		
		}


		/*************************************************************************************
		Called to update the tab areas
		*************************************************************************************/
		protected virtual void UpdateAreas(int nMaxWidth)
		{
			if( nMaxWidth!=-1 ) m_nMaxWidth = nMaxWidth;

			// Create a graphics object to messure the text
			Bitmap oBit = new Bitmap(1,1);
			System.Drawing.Graphics g = System.Drawing.Graphics.FromImage(oBit);

			// Slope and border space
			int nSpace = (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) + (m_nBorderSpace*2);

			// Get the text width
			int nTxWidth = Convert.ToInt32(g.MeasureString(m_sText, (m_Font==null ? m_oParent.Font:m_Font)).Width);

			// Does the tab have a valid image
			bool bHasImg = (m_nImage>=0 && m_oParent.m_ImageList!=null) || (m_nHotImage>=0 && (m_oParent.m_ImageList!=null || m_oParent.m_HotImageList!=null) );
	
			// Get the image list information
			int nImgHeight = 0, nImgWidth = 0;
			if( bHasImg ) 
			{
				ImageList oImageList = (m_oParent.m_ImageList==null ? m_oParent.m_HotImageList:m_oParent.m_ImageList);
				nImgHeight = oImageList.ImageSize.Height;
				nImgWidth = oImageList.ImageSize.Width;
			}

			//  Is the image horizontally centered
			bool bImgCentered = (m_nImgAlign==ContentAlignment.TopCenter || m_nImgAlign==ContentAlignment.MiddleCenter || m_nImgAlign==ContentAlignment.BottomCenter);

			// If the image is horizontally centered make the text width the max value of the image width and text width
			if( bHasImg && bImgCentered ) nTxWidth = Math.Max(nImgWidth, nTxWidth);

			// Update the minimum width
			m_nMinWidth = nSpace + (bHasImg ? nImgWidth:5);

			// Update the tab width
			if( m_nMaxWidth==-1 && m_bAutoWidth ) 
			{
				// No width was specified and auto width is on; calculate the width
				m_nRealWidth = nSpace + nTxWidth + (bHasImg ? nImgWidth+(bImgCentered ? 0:(int)m_nImgSpace):0);
				if( m_bAutoWidth ) m_nWidth = m_nRealWidth;
			}
			else
			{
				// The real width is the minimum width or the specified with or the tab width property
				m_nRealWidth = Math.Max(m_nMinWidth, (m_nMaxWidth==-1 ? m_nWidth:m_nMaxWidth));
			}
	
			// Image area
			if( bHasImg )
			{
				// Vertical position
				switch( m_nImgAlign )
				{
					case ContentAlignment.TopLeft:
						m_ImgArea.Y = 0;
						break;
					case ContentAlignment.TopCenter:
						m_ImgArea.Y = 0;
						break;
					case ContentAlignment.TopRight:
						m_ImgArea.Y = 0;
						break;
					/*case ContentAlignment.Top:
						m_ImgArea.Y = 0;
						break;*/
					case ContentAlignment.MiddleLeft:
						m_ImgArea.Y = m_nHeight/2 - nImgHeight/2;
						break;
					case ContentAlignment.MiddleCenter:
						m_ImgArea.Y = m_nHeight/2 - nImgHeight/2;
						break;
					case ContentAlignment.MiddleRight:
						m_ImgArea.Y = m_nHeight/2 - nImgHeight/2;
						break;
					/*case ContentAlignment.Middle:
						m_ImgArea.Y = m_nHeight/2 - nImgHeight/2;
						break;*/
					case ContentAlignment.BottomLeft:
						m_ImgArea.Y = m_nHeight - nImgHeight;
						break;
					case ContentAlignment.BottomCenter:
						m_ImgArea.Y = m_nHeight - nImgHeight;
						break;
					case ContentAlignment.BottomRight:
						m_ImgArea.Y = m_nHeight - nImgHeight;
						break;
					/*case ContentAlignment.Bottom:
						m_ImgArea.Y = m_nHeight - nImgHeight;
						break;*/
				}
				// Horizontal position
				switch( m_nImgAlign )
				{
					case ContentAlignment.TopLeft:
						m_ImgArea.X = m_nBorderSpace + (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2);
						break;
					case ContentAlignment.MiddleLeft:
						m_ImgArea.X = m_nBorderSpace + (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2);
						break;
					case ContentAlignment.BottomLeft:
						m_ImgArea.X = m_nBorderSpace + (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2);
						break;
					/*case ContentAlignment.Left:
						m_ImgArea.X = m_nBorderSpace + (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2);
						break;*/
					case ContentAlignment.TopCenter:
						m_ImgArea.X = m_nRealWidth/2-nImgWidth/2;
						break;
					case ContentAlignment.MiddleCenter:
						m_ImgArea.X = m_nRealWidth/2-nImgWidth/2;
						break;
					case ContentAlignment.BottomCenter:
						m_ImgArea.X = m_nRealWidth/2-nImgWidth/2;
						break;
					/*case ContentAlignment.Center:
						m_ImgArea.X = m_nRealWidth/2-nImgWidth/2;
						break;*/
					case ContentAlignment.TopRight:
						m_ImgArea.X = (m_nWidth-m_nBorderSpace-nImgWidth - (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2));
						break;
					case ContentAlignment.MiddleRight:
						m_ImgArea.X = (m_nWidth-m_nBorderSpace-nImgWidth - (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2));
						break;
					case ContentAlignment.BottomRight:
						m_ImgArea.X = (m_nWidth-m_nBorderSpace-nImgWidth - (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)+2));
						break;
				}

				// Size
				m_ImgArea.Width = nImgWidth;
				m_ImgArea.Height = nImgHeight;
	
			}
			else 
			{
				// No image; clear the area
				m_ImgArea.X=0;
				m_ImgArea.Y=0;
				m_ImgArea.Width=0;
				m_ImgArea.Height=0;
			}

			// Text area
			if( !bHasImg || m_nImgAlign==ContentAlignment.MiddleCenter)// || m_nImgAlign==ContentAlignment.Center )
			{
				m_nSlopePhoto = (m_nStyle == TabStyle.Photoshop) ? (short)0 : m_nSlope;
				// Area covers the entire tab
				m_TxArea.X = (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlopePhoto)+2) + m_nBorderSpace;
				m_TxArea.Width = m_nRealWidth - (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) - (m_nBorderSpace*2);
				m_TxArea.Y = 0;
				m_TxArea.Height = m_nHeight;
			}
			else
			{

				switch( m_nImgAlign )
				{
					case ContentAlignment.TopLeft:
					case ContentAlignment.MiddleLeft:
					case ContentAlignment.BottomLeft:
					//case ContentAlignment.Left:
						// Cover the middle and right
						m_TxArea.X = m_ImgArea.Right + m_nImgSpace;
						m_TxArea.Width = m_nRealWidth - (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) - (m_nBorderSpace*2) - nImgWidth - m_nImgSpace;
						m_TxArea.Y = 0;
						m_TxArea.Height = m_nHeight;
						break;
					case ContentAlignment.TopCenter:
						// Covers the bottom
						m_nSlopePhoto = (m_nStyle == TabStyle.Photoshop) ? (short)0 : m_nSlope;
						m_TxArea.X = (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlopePhoto)+2) + m_nBorderSpace;
						m_TxArea.Width = m_nRealWidth - (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) - (m_nBorderSpace*2);
						m_TxArea.Y = m_ImgArea.Bottom + m_nImgSpace;
						m_TxArea.Height = m_nHeight - m_ImgArea.Height - m_nImgSpace;
						break;
					case ContentAlignment.BottomCenter:
						// Covers the top
						m_nSlopePhoto = (m_nStyle == TabStyle.Photoshop) ? (short)0 : m_nSlope;
						m_TxArea.X = (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlopePhoto)+2) + m_nBorderSpace;
						m_TxArea.Width = m_nRealWidth - (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) - (m_nBorderSpace*2);
						m_TxArea.Y = 0;
						m_TxArea.Height = m_nHeight - m_ImgArea.Height - m_nImgSpace;
						break;
					case ContentAlignment.TopRight:
					case ContentAlignment.MiddleRight:
					case ContentAlignment.BottomRight:
					//case ContentAlignment.Right:
						// Covers the middle and left
						m_nSlopePhoto = (m_nStyle == TabStyle.Photoshop) ? (short)0 : m_nSlope;
						m_TxArea.X = (m_bOverlaySlope ? 0:(int)Math.Abs(m_nSlope)) + m_nBorderSpace;
						m_TxArea.Width = m_nRealWidth - (m_bOverlaySlope ? 0:((Math.Abs(m_nSlope)+2)*2)) - (m_nBorderSpace*2) - nImgWidth - m_nImgSpace;
						m_TxArea.Y = 0;
						m_TxArea.Height = m_nHeight;
						break;
				}

				// If the width or height is negative don't draw text
				if( m_TxArea.Width<0 || m_TxArea.Height<0 ) 
				{
					m_TxArea.X = 0;
					m_TxArea.Y = 0;
					m_TxArea.Width = 0;
					m_TxArea.Height = 0;
				}
			}

			// Clean up
			g.Dispose();
			oBit.Dispose();

			m_bSizeChanged = false;
		}


		/*************************************************************************************
		Called by the tab control to update internal variables
		*************************************************************************************/
		internal virtual void Update(int X, int Y, int nHeight, int nMaxWidth, TabState nState)
		{
			// Don't update if nothing changed
			if( m_nX==X && m_nY==Y && m_nHeight==nHeight ) return;
	
			// Area variables
			m_nX = X;
			m_nY = Y;
			if( m_bSizeChanged || nMaxWidth!=m_nMaxWidth || m_nHeight!=nHeight ) 
			{
				m_nHeight = nHeight;		
				UpdateAreas(nMaxWidth);
				m_nMaxWidth = nMaxWidth;
			}

			// Update the coordinates
			UpdateCoordinates();

			// Update the drawing region
			if( m_oRgn!=null ) m_oRgn.Dispose();
			byte[] aPathType = new byte[] { (byte)PathPointType.Start, (byte)PathPointType.Line, (byte)PathPointType.Line, (byte)PathPointType.Line };
			m_aPoints[1].Offset(-1,0); m_aPoints[2].Offset(-2,0);	// This ensures that no part region is outside the tab area
			m_oRgn = new Region( new GraphicsPath(m_aPoints, aPathType) );
			m_aPoints[1].Offset(1,0); 
			m_aPoints[2].Offset(2,0);		// Return the coodinates to default

			// Update the state
			m_nState = nState;
		}


		/*************************************************************************************
		Updates the tab state
		*************************************************************************************/
		internal virtual void UpdateState(TabState nState)
		{
			// Exit if the specified state is already set
			if( m_nState==nState ) return;

			// Clean up tabs that are no longer visible
			if( (nState & TabState.Invisible)==TabState.Invisible && !IsInvisible ) DisposeVisible();

			// New state
			m_nState = nState;
		}
		internal void EnableState(TabState nState)
		{
			UpdateState(m_nState | nState);
		}
		internal void DisableState(TabState nState)
		{
			UpdateState(m_nState & ~nState);
		}


		/*************************************************************************************
		Draws the tab
		*************************************************************************************/
		internal virtual void DrawTab(System.Drawing.Graphics g)
		{
			// If no region is available exit
			if( m_oRgn==null ) return;

			// Draw the background
			OnDrawBackground(g);

			// Clipping region
			g.Clip = m_oRgn;

			// Draw the text
			if( !m_TxArea.IsEmpty ) OnDrawText(g);

			// Draw the image
			OnDrawImage(g);

			// Clear the clipping region
			g.ResetClip();

			// Draw the border
			OnDrawBorders(g);
		}


		/*************************************************************************************
		Releases memory used by visible tabs when no longer needed
		*************************************************************************************/
		protected virtual void DisposeVisible()
		{
			if( m_oRgn!=null ) m_oRgn.Dispose();
			m_oRgn = null;
			m_aPoints = null;
			m_nX = 0;
			m_nY = 0;
		}


		/*************************************************************************************
		Updates the tab coordinates array
		*************************************************************************************/
		protected virtual void UpdateCoordinates()
		{
			Point tLeftTop, tRightTop, tLeftBottom, tRightBottom;

			int nSlope = m_nSlope * (m_nDirection==TabDirection.Top? 1:-1);

			// Direction of the slope
			bool bDown = nSlope>=0;

			// Region coordinates
			if( bDown )	
			{
				tLeftTop = new Point(m_nX, m_nY);
				tRightTop = new Point(m_nX+m_nRealWidth, m_nY);
				int m_nSlopePhoto1 = (m_nStyle == TabStyle.Photoshop) ? 0 : nSlope;
				tRightBottom = new Point(m_nX+m_nRealWidth-m_nSlopePhoto1, m_nY+m_nHeight);
				tLeftBottom = new Point(m_nX+nSlope, m_nY+m_nHeight);
			}
			else
			{
				int m_nSlopePhoto1 = (m_nStyle == TabStyle.Photoshop) ? 0 : nSlope;
				tLeftTop = new Point(m_nX-m_nSlopePhoto1, m_nY);
				tRightTop = new Point(m_nX+m_nRealWidth+nSlope, m_nY);
				tRightBottom = new Point(m_nX+m_nRealWidth, m_nY+m_nHeight);
				tLeftBottom = new Point(m_nX, m_nY+m_nHeight);
			}
	
			m_aPoints = new Point[] { tLeftTop, tRightTop, tRightBottom, tLeftBottom };
		}


		/*************************************************************************************
		Draws the image
		*************************************************************************************/
		protected virtual void OnDrawImage(System.Drawing.Graphics g)
		{
			if( m_oParent==null ) return;
			int nIndex = (IsHot ? m_nHotImage:m_nImage);
			if( nIndex==-1 ) return;

			// Get the image list
			ImageList oImageList = (IsHot && m_oParent.m_HotImageList!=null ? m_oParent.m_HotImageList:m_oParent.m_ImageList);
			if( oImageList==null ) return;

			// Draw the image
			m_ImgArea.Offset(m_nX, m_nY);
			try { oImageList.Draw(g, m_ImgArea.X, m_ImgArea.Y, nIndex); } 
			catch {}
			m_ImgArea.Offset(m_nX*-1, m_nY*-1);
		}


		/*************************************************************************************
		Draws the tab background
		*************************************************************************************/
		protected virtual void OnDrawBackground(System.Drawing.Graphics g)
		{
			Color nColor = ( IsActive ? m_ActiveBackColor : (IsHot ? m_HotBackColor : m_BackColor));
			g.FillRegion( new SolidBrush(nColor), m_oRgn );		
		}


		/*************************************************************************************
		Draws the tab text
		*************************************************************************************/
		protected virtual void OnDrawText(System.Drawing.Graphics g)
		{
			// Font; if the control is active and hot the hot font will be used
			Font oFont = (IsHot && m_HotFont!=null ? m_HotFont : (IsActive && m_ActiveFont!=null ? m_ActiveFont : m_Font));
			if( oFont==null && m_oParent==null ) return;
			if( oFont==null ) oFont = m_oParent.Font;

			Color nColor = (IsHot ? m_HotForeColor : (IsActive ? m_ActiveForeColor : m_ForeColor));

			// Draw
			m_TxArea.Offset(m_nX, m_nY);
			g.DrawString(m_sText, oFont, new SolidBrush(nColor), m_TxArea, m_TextFormat);
			m_TxArea.Offset(m_nX*-1, m_nY*-1);
		}


		/*************************************************************************************
		Draws the tab borders
		*************************************************************************************/
		protected virtual void OnDrawBorders(System.Drawing.Graphics g)
		{
			// Create a pen used to draw the borders
			Pen oPen = new Pen(Color.Black, 1);

			// Is the tab active?
			bool bActive = IsActive;

			TabBorders nBr = (bActive ? m_nABorderSides:m_nBorderSides);

			// Top outer border
			if( (nBr & TabBorders.TopOuter) == TabBorders.TopOuter )
			{
				oPen.Color = (bActive ? m_ATOColor:m_TOColor);
				m_aPoints[1].Offset(-1,0);
				g.DrawLine(oPen, m_aPoints[0], m_aPoints[1]);
				m_aPoints[1].Offset(1,0);
			}
			// Top inner border
			if( (nBr & TabBorders.TopInner) == TabBorders.TopInner )
			{
				oPen.Color = (bActive ? m_ATIColor:m_TIColor);
				m_aPoints[0].Offset(0,1);
				m_aPoints[1].Offset(-1,1);
				g.DrawLine(oPen, m_aPoints[0], m_aPoints[1]);
				m_aPoints[0].Offset(0,-1); 
				m_aPoints[1].Offset(1,-1);
			}
			// Bottom outer border
			if( (nBr & TabBorders.BottomOuter) == TabBorders.BottomOuter )
			{
				oPen.Color = (bActive ? m_ABOColor:m_BOColor);
				m_aPoints[3].Offset(0,-1);
				m_aPoints[2].Offset(-1,-1);
				g.DrawLine(oPen, m_aPoints[3], m_aPoints[2]);
				m_aPoints[3].Offset(0,1); 
				m_aPoints[2].Offset(1,1);
			}
			// Bottom inner border
			if( (nBr & TabBorders.BottomInner) == TabBorders.BottomInner )
			{
				oPen.Color = (bActive ? m_ABIColor:m_BIColor);
				m_aPoints[3].Offset(0,-2);
				m_aPoints[2].Offset(-1,-2);
				g.DrawLine(oPen, m_aPoints[3], m_aPoints[2]);
				m_aPoints[3].Offset(0,2); 
				m_aPoints[2].Offset(1,2);
			}
			// Left outer border
			if( (nBr & TabBorders.LeftOuter) == TabBorders.LeftOuter )
			{
				oPen.Color = (bActive ? m_ALOColor:m_LOColor);
				m_aPoints[3].Offset(0,-1);
				g.DrawLine(oPen, m_aPoints[0], m_aPoints[3]);
				m_aPoints[3].Offset(0,1);
			}
			// Left inner border
			if( (nBr & TabBorders.LeftInner) == TabBorders.LeftInner )
			{
				oPen.Color = (bActive ? m_ALIColor:m_LIColor);
				m_aPoints[0].Offset(1,0); 
				m_aPoints[3].Offset(1,-1);
				g.DrawLine(oPen, m_aPoints[0], m_aPoints[3]);
				m_aPoints[0].Offset(-1,0); 
				m_aPoints[3].Offset(-1,1);
			}
			// Right outer border
			if( (nBr & TabBorders.RightOuter) == TabBorders.RightOuter )
			{
				oPen.Color = (bActive ? m_AROColor:m_ROColor);
				m_aPoints[1].Offset(-1,0);
				m_aPoints[2].Offset(-1,-1);
				g.DrawLine(oPen, m_aPoints[1], m_aPoints[2]);
				m_aPoints[1].Offset(1,0);
				m_aPoints[2].Offset(1,1);
			}
			// Right inner border
			if( (nBr & TabBorders.RightInner) == TabBorders.RightInner )
			{
				oPen.Color = (bActive ? m_ARIColor:m_RIColor);
				m_aPoints[1].Offset(-2,0);
				m_aPoints[2].Offset(-2,-1);
				g.DrawLine(oPen, m_aPoints[1], m_aPoints[2]);
				m_aPoints[1].Offset(2,0); 
				m_aPoints[2].Offset(2,1);
			}

			oPen.Dispose();
		}


		/*************************************************************************************
		Updates the tab style
		*************************************************************************************/
		protected void UpdateStyle()
		{
			if( m_nStyle==TabStyle.None ) return;

			// VisualStudio; This is the default style
			if( m_nStyle==TabStyle.VisualStudio )
			{
				m_nSlope			= 0;
				m_LOColor			= Color.Transparent;
				m_LIColor			= Color.Transparent;
				m_TOColor			= Color.Transparent;
				m_TIColor			= Color.Transparent;
				m_ROColor			= Color.DarkGray;
				m_RIColor			= Color.Transparent;
				m_BOColor			= Color.Transparent;
				m_BIColor			= Color.Transparent;
				m_ALOColor			= Color.White;
				m_ALIColor			= Color.Transparent;
				m_ATOColor			= Color.Transparent;
				m_ATIColor			= Color.Transparent;
				m_AROColor			= Color.Black;
				m_ARIColor			= Color.Transparent;
				m_ABOColor			= Color.Black;
				m_ABIColor			= Color.FromArgb(247, 243, 233);
				m_BackColor			= Color.FromArgb(247, 243, 233);
				m_HotBackColor		= Color.FromArgb(247, 243, 233);
				m_ActiveBackColor	= SystemColors.Control;
				m_ForeColor			= Color.DarkGray;
				m_HotForeColor		= Color.Blue;
				m_ActiveForeColor	= SystemColors.ControlText;
				m_nBorderSides		= TabBorders.None;
				m_nABorderSides		= TabBorders.All;
	
				// Excel or InterDev
			}
			else if( m_nStyle==TabStyle.Excel || m_nStyle==TabStyle.InterDev )
			{
				m_nSlope			= 15;
				m_LOColor			= Color.Black;
				m_LIColor			= Color.White;
				m_TOColor			= Color.Transparent;
				m_TIColor			= Color.Transparent;
				m_ROColor			= Color.Black;
				m_RIColor			= (m_nStyle==TabStyle.Excel ? Color.DarkGray:Color.DarkGray);
				m_BOColor			= (m_nStyle==TabStyle.Excel ? Color.DarkGray:Color.Black);
				m_BIColor			= Color.Transparent;
				m_ALOColor			= Color.Black;
				m_ALIColor			= Color.White;
				m_ATOColor			= Color.White;
				m_ATIColor			= Color.White;
				m_AROColor			= Color.Black;
				m_ARIColor			= Color.White;
				m_ABOColor			= Color.DarkGray;
				m_ABIColor			= Color.Transparent;
				m_BackColor			= (m_nStyle==TabStyle.Excel ? SystemColors.Control:Color.DarkGray);
				m_HotBackColor		= (m_nStyle==TabStyle.Excel ? SystemColors.Control:Color.DarkGray);
				m_ActiveBackColor	= Color.White;
				m_ForeColor			= (m_nStyle==TabStyle.Excel ? SystemColors.ControlText:Color.White);
				m_HotForeColor		= (m_nStyle==TabStyle.Excel ? Color.Blue:Color.Black);
				m_ActiveForeColor	= SystemColors.ControlText;
				m_nBorderSides		= TabBorders.All;
				m_nABorderSides		= TabBorders.All;
			}
			else if (m_nStyle == TabStyle.Photoshop)
			{
				m_ABOColor			= Color.Black;
				m_ABIColor			= Color.Transparent;
				m_BOColor			= Color.DarkGray;
				m_BIColor			= Color.Transparent;
				m_ATOColor			= Color.Transparent;
				m_ATIColor			= Color.Transparent;
				m_TOColor			= Color.Transparent;
				m_TIColor			= Color.Transparent;
			}

			// Update the direction
			m_nDirection = TabDirection.Top;
			UpdateDirection();
		}

		/*************************************************************************************
		Changes the tab direction
		*************************************************************************************/
		internal void UpdateDirection()
		{
			if( m_oParent==null ) return;
			if( m_nDirection==m_oParent.m_nDirection ) return;

			m_nDirection = m_oParent.m_nDirection;

			// Flip the top and bottom colors
			Color nColor = m_ABIColor;
			m_ABIColor = m_ATIColor;
			m_ATIColor = nColor;
			nColor = m_ABOColor;
			m_ABOColor = m_ATOColor;
			m_ATOColor = nColor;
			nColor = m_BIColor;
			m_BIColor = m_TIColor;
			m_TIColor = nColor;
			nColor = m_BOColor;
			m_BOColor = m_TOColor;
			m_TOColor = nColor;

			if( m_nStyle==TabStyle.VisualStudio ) 
			{
				if( m_nDirection==TabDirection.Top ) 
				{
					m_ATOColor			= Color.Transparent;
					m_ATIColor			= Color.Transparent;
				} 
				else 
				{
					m_ATOColor = Color.White;
					m_ATIColor = Color.Transparent;				
				}
			}
		}
	}



	/*****************************************************************************************
	Tabs collection object
	*****************************************************************************************/
	public class Tabs : BaseCollection
	{
		/*************************************************************************************
		Variables Declaration
		*************************************************************************************/
		internal ArrayList m_List;
		TabCtl m_oParent;


		/*************************************************************************************
		Constructor
		*************************************************************************************/
		internal Tabs(TabCtl oParent) 
		{
			m_List = new ArrayList();
			m_oParent = oParent;
			if( m_oParent==null ) throw( new Exception("Invalid parent control") ); 
		}


		/*************************************************************************************
		Finalize
		*************************************************************************************/
		protected void Finalize()
		{
			m_oParent = null;
			Clear();

			//base.Finalize();
		}


		/*************************************************************************************
		Adds a new tab to the collection
		*************************************************************************************/
		[Description("Adds the specified tab to the collection."),]
		public int Add(Tab oItem)
		{
			if( oItem==null ) return -1;
			SetTabParent(ref oItem);
			oItem.m_nIndex = m_List.Count;
			int nReturn = m_List.Add(oItem);
			m_oParent.TabsChanged();

			return nReturn;
		}
		[Description("Appends a new tab to the collection."),]
		public Tab Append() 
		{
			Tab oTab = new Tab();
			SetTabParent(ref oTab);
			oTab.m_nIndex = m_List.Count;
			m_List.Add(oTab);		
			m_oParent.TabsChanged();
			return oTab;
		}


		/*************************************************************************************
		Returns a item at the specified position
		*************************************************************************************/
		public Tab this[int nIndex] { get { return (Tab)m_List[nIndex]; } }


		/*************************************************************************************
		Inserts a tab at the specified position
		*************************************************************************************/
		[Description("Inserts an tab into the collection at the specified index."),]
		public int Insert(int nIndex, Tab oItem) 
		{
			if( oItem==null || nIndex<0 || nIndex>m_List.Count ) return -1;
	
			if( nIndex==Count ) return Add(oItem);
			SetTabParent(ref oItem);
			m_List.Insert(nIndex, oItem);
			Reindex(nIndex-1);
			m_oParent.TabsChanged();
			return nIndex;
		}


		/*************************************************************************************
		Removes tabs from the collection
		*************************************************************************************/
		[Description("Removes a specific tab from the collection."),]
		public void Remove(Tab oItem) 
		{
			if( oItem.m_oParent!=m_oParent ) throw( new Exception("Tab is not a member of this collection.") );

			m_List.Remove(oItem);
			oItem.m_nIndex = -1;
			oItem.m_oParent = null;
			Reindex(oItem.m_nIndex);
			m_oParent.TabsChanged();
		}
		[Description("Removes the tab at the specified index of the collection."),]
		public void RemoveAt(int nIndex)
		{
			if( nIndex<0 || nIndex>=m_List.Count ) throw( new Exception("Invalid tab index.") );
	
			Tab oTab = (Tab)m_List[nIndex];
			m_List.RemoveAt(nIndex); 
			oTab.m_nIndex = -1;
			oTab.m_oParent = null;
			Reindex(oTab.m_nIndex);
			m_oParent.TabsChanged();
		}
		[Description("Clears the contents of the collection."),]
		public void Clear() 
		{ 
			foreach( Tab oTab in m_List ) 
			{ 
				oTab.m_nIndex = -1; 
				oTab.m_oParent = null; 
			}

			m_List.Clear();
			if( m_oParent!=null ) m_oParent.TabsChanged();
		}


		/*************************************************************************************
		Updates the index of all tabs in the specified range
		*************************************************************************************/
		internal void Reindex(int nStart, int nEnd)
		{
			// Validate the range
			if( nStart==-1 ) nStart = 0;
			if( nEnd==-1 ) nEnd = m_List.Count-1;
			if( nStart>=m_List.Count || nEnd>=m_List.Count || nStart>nEnd ) return;

			// Update the indexes
			for( int nLoop=nStart; nLoop<=nEnd; nLoop++ ) ((Tab)m_List[nLoop]).m_nIndex = nLoop;
		}
		internal void Reindex(int nStart)
		{
			Reindex(nStart, -1);
		}
		internal void Reindex()
		{
			Reindex(-1, -1);
		}


		/*************************************************************************************
		Update the parent of a tab
		*************************************************************************************/
		internal void SetTabParent(ref Tab oTab)
		{
			// Throw an error if the tab already has a parent
			if( oTab.m_oParent!=null && m_oParent!=null ) throw( new Exception("Tab already is already associated with a control.") );

			oTab.m_oParent = m_oParent;
			if( oTab.Style==TabStyle.None ) oTab.Style = m_oParent.m_nTabStyle;
			oTab.UpdateDirection();
		}


		/*************************************************************************************
		Changes a object index
		*************************************************************************************/
		internal void SetObjectIndex(object oItem, int nIndex)
		{
			Tab oTab = oItem as Tab;

			// Validate the index
			if( nIndex<0 ) nIndex = 0;
			if( nIndex>=m_List.Count ) nIndex = m_List.Count-1;
			if( oTab.m_nIndex==nIndex ) return;	// Exit if the index is not changing
	
			int	nReindexStart = Math.Min(oTab.m_nIndex, nIndex);
	
			// Change the tab position
			m_List.RemoveAt(oTab.m_nIndex);
			m_List.Insert(nIndex, oItem);

			Reindex(nReindexStart);

			// Update the parent
			m_oParent.TabsChanged();
		}


		/*************************************************************************************
		List property Override
		*************************************************************************************/
		protected override ArrayList List 
		{ 
			get { return m_List; } 
		}


		/*************************************************************************************
		All property
		*************************************************************************************/
		[DefaultValue(null), Browsable(false), RefreshProperties(RefreshProperties.All) ]//Persistable(PersistableSupport.All), 	
		public Tab[] All 
		{
			get 
			{		
				Tab[] oTemp = new Tab[m_List.Count];
				m_List.CopyTo(oTemp, 0);
				return oTemp;
			}
			set 
			{
				Clear();
				int nIndex = 0;
				Tab tmpTab;
				
				foreach (Tab oTab in value)
				{
					tmpTab = oTab;
					SetTabParent(ref tmpTab);
					tmpTab.m_nIndex = nIndex;
					m_List.Add(tmpTab);
					nIndex++;
				}
			
				m_oParent.TabsChanged();
			}
		}

		/*************************************************************************************
		Updates the tabs styles
		*************************************************************************************/
		internal void UpdateStyle(TabStyle nStyle)
		{
			foreach( Tab oTab in m_List ) oTab.Style = nStyle;
		}

		/*************************************************************************************
		Updates the tabs direction
		*************************************************************************************/
		internal void UpdateDirection()
		{
			foreach( Tab oTab in m_List ) oTab.UpdateDirection();
		}

		/*************************************************************************************
		Returns a object index
		*************************************************************************************/
		[Description("Searches the collection for the specified tab and returns the zero-based index of the first accurrence."),]
		public int IndexOf(Tab oTab)  
		{
			return m_List.IndexOf(oTab);
		}
	}


	/*****************************************************************************************
	Tabs collection editor support
	*****************************************************************************************/
	public class TabsCollectionEditor : System.ComponentModel.Design.CollectionEditor
	{
		public TabsCollectionEditor(Type type) : base(type) {}

		protected override object CreateInstance(Type itemType) 
		{
			object obj = base.CreateInstance(itemType);
		
			if(obj is Tab && this.Context!=null && Context.Instance!=null) 
			{ 
				((TabCtl)Context.Instance).Tabs.Add((Tab)obj);
			}
			return obj;
		}
	}

}

