using System;
using System.Collections;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Data;
using System.Windows.Forms;
	

namespace MyTabControl
{
	public class TabChangeEventArgs : EventArgs 
	{
		private Tab m_oPrevTab;
		private Tab m_oTab;
		bool m_bCancel = false;

		internal TabChangeEventArgs(Tab oPrevTab, Tab oTab) 
		{
			m_oTab = oTab;
			m_oPrevTab = oPrevTab;
		}
		public Tab Tab { get{ return m_oTab; } }
		public Tab PrevTab { get{ return m_oPrevTab; } }
		public bool Cancel { get{ return m_bCancel; }
			set{ m_bCancel = value; }
		}
	}

	public class SBClickEventArgs : EventArgs 
	{
		private SB m_nSB;
		bool m_bCancel = false;
		internal SBClickEventArgs(SB nSB) 
		{
			m_nSB = nSB;
		}
		public SB Button { get{ return m_nSB; } }
		public bool Cancel { get{ return m_bCancel; }
			set{ m_bCancel = value; }
		}
	}
	// Event delegates
	public delegate void TabChangeEventHandler(object sender, TabChangeEventArgs e);
	public delegate void SBClickEventHandler(object sender, SBClickEventArgs e);
	

	
	/*****************************************************************************************
	Enumerators
	*****************************************************************************************/
	public enum SB { First=2, Prev=4, Next=8, Last=16, All=30, FirstLast=18, PrevNext=12, None=0 }
	public enum SBPos { Left=0, Right }
	public enum TabDirection { Top=0, Bottom }



	/*****************************************************************************************
	Tab Control
	*****************************************************************************************/
	[DefaultEvent("TabChange"),]
	public class TabCtl : System.Windows.Forms.Control
	{
		/*************************************************************************************
		Variables declaration
		*************************************************************************************/

		// Properties
		protected short			m_nTopMargin		= 0,
			m_nLeftMargin		= 0,
			m_nRightMargin		= 0,
			m_nBottomMargin		= 2,
			m_nHorLineWidth		= 1,
			m_nTabSpacing		= 0,
			m_nInitTabSpace		= 3;	
		protected bool			/*m_bAutoSize			= false,*/
			m_bAutoUpdate		= true,
			m_bHorLine			= true,
			m_bReverseZOrder	= true,
			draggedTab			= false,
			leaveTab			= true;
		protected Color			m_HorLineColor		= Color.Black,
			m_SBHotFore			= Color.Blue,
			m_SBFore			= Color.Black,
			m_SBDisableFore		= Color.DarkGray,
			m_SBHotBack			= SystemColors.Control,
			m_SBBack			= SystemColors.Control,
			m_SBDisableBack		= SystemColors.Control,
			m_SepColor			= Color.DarkGray;
		protected short			m_nSBWidth			= (short)SystemInformation.VerticalScrollBarWidth;
		protected SB			m_nSBVisible		= SB.All;
		protected Border3DStyle	m_SBBorder			= Border3DStyle.Adjust,
			m_SBABorder			= Border3DStyle.SunkenOuter,
			m_SBHBorder			= Border3DStyle.RaisedInner,
			m_Border			= Border3DStyle.SunkenOuter;
		protected Tabs			m_oTabs;
		protected Cursor		m_HotCursor			= Cursors.Hand,
			m_Cursor			= Cursors.Default,
			m_SBCursor			= Cursors.Hand;
		protected MyTabControl.Graphics.GraphicsBuffer m_oBuffer			= null;
		protected SBPos			m_nSBPosition		= SBPos.Left;
		internal TabStyle		m_nTabStyle			= TabStyle.VisualStudio;
		internal ImageList		m_ImageList			= null,
			m_HotImageList		= null;
		internal TabDirection	m_nDirection		= TabDirection.Top;

		// Internal variables
		protected Rectangle		m_TabsArea,							// Tabs area
			m_SBArea;							// Scroll buttons area
		protected ArrayList		m_aVisible;							// Visible tabs array
		protected int			m_nVisibleCount,					// Number of visible tabs
			m_nFirstIndex,						// Index of first visible tab
			m_nLastIndex;						// Index of last visible tab
		protected bool			m_bLastFullyVisible = false,		// True if the last visible tab is fully visible
			m_bFirstVisible		= false,		// True if the first tab is visile
			m_bLastVisible		= false;		// True if the last tab is visible
		protected Tab			m_oHTab				= null,			// Hot tab
			m_oATab				= null,			// Selected tab
			m_oDTab				= null;			// Dragged tab
		protected SB	m_nSB,				 				// Hot scroll button
			m_nDSB;								// Mouse down scroll button
		protected Pen			m_oSepPen;							// Separator pen


		// Events
		public event SBClickEventHandler SBClick;
		public event TabChangeEventHandler TabChange;

		/*************************************************************************************
		Events implementation
		*************************************************************************************/
		protected virtual bool OnTabChange(Tab oTab, Tab oPrevTab) 
		{ 
			// Raises the tab change event; returns true to cancel, false to resume		
			if( TabChange==null ) return false;
			TabChangeEventArgs oArg = new TabChangeEventArgs(oTab, oPrevTab);
			this.TabChange(this, oArg); 
			return oArg.Cancel;
		}
		protected virtual bool OnSBClick(SB nSB) 
		{ 
			// Raises the SB click event; returns true to cancel, false to resume		
			if( SBClick==null ) return false;	
			SBClickEventArgs oArg = new SBClickEventArgs(nSB);
			this.SBClick(this, oArg); 
			return oArg.Cancel;
		}


		/*************************************************************************************
		Constructors
		*************************************************************************************/
		public TabCtl() : base() 
		{
			// Control styles
			SetStyle(ControlStyles.ResizeRedraw|ControlStyles.Opaque, true);
			SetStyle(ControlStyles.Selectable, false);

			// Initialize objects
			m_aVisible = new ArrayList();
			m_oTabs = new Tabs(this);
			m_oSepPen = new Pen(Color.DarkGray, 1);
			m_oBuffer = new MyTabControl.Graphics.GraphicsBuffer();
			BackColor = Color.FromArgb(247, 243, 233);

			// Control style flags
			SetStyle(ControlStyles.Opaque|ControlStyles.ResizeRedraw, true);
			SetStyle(ControlStyles.Selectable, false);
			AllowDrop = true;
			/*Panel cont = new Panel();
			cont.AutoScroll = true;
			cont.Size = new Size(200, 200);
			cont.Dock = DockStyle.Bottom;*/
			//cont.BorderStyle = BorderStyle.RaisedInner;
			//this.Controls.Add(cont);
		}


		/*************************************************************************************
		OnPaint Override
		*************************************************************************************/
		protected override void OnPaint(PaintEventArgs e) 
		{
			// Get the drawing graphics object
			System.Drawing.Graphics g;
			if( m_oBuffer!=null )	
				g = m_oBuffer.RequestBuffer(e.Graphics, ClientRectangle.Width, ClientRectangle.Height);
			else					
				g = e.Graphics;

			// Control background
			OnDrawBackground(g);

			// Draw the horizontal line
			if( m_bHorLine ) OnDrawHorizontalLine(g);

			// Draw the visible tabs
			int nNoSep = (m_oATab==null ? -1: GetPrevVisible(m_oATab.Index-1));
			foreach(Tab oTab in m_aVisible)	
			{
				// Don't draw the active tab; we want the active tab to be on top off all others
				if( oTab!=m_oATab ) 
				{
					oTab.DrawTab(g);

					// If draw separators is true draw separators for all tabs with 
					// VisualStudio style except the one prior to the active
					if( m_oSepPen!=null && nNoSep!=oTab.Index && oTab.Style==TabStyle.VisualStudio ) OnDrawSeparator(g, oTab);				
				}
			}
			if( m_oATab!=null ) m_oATab.DrawTab(g);

			// Draw the scroll buttons
			OnDraAllSB(g);

			// Draws the control border
			OnDrawBorder(g);

			// Transfer the contents of the double buffer object to the destination object
			if( g!=e.Graphics ) m_oBuffer.PaintBuffer(e.Graphics, 0, 0);
		}


		/*************************************************************************************
		Control is resizing
		*************************************************************************************/
		protected override void OnResize(System.EventArgs e)
		{
			base.OnResize(e);

			// Update the areas
			UpdateAreas();

			// Update the control
			try { EnsureFirst(m_nFirstIndex); } 
			catch {}
		}


		/*************************************************************************************
		OnMouseMove Override
		*************************************************************************************/
		protected override void OnMouseMove(MouseEventArgs e)
		{
			base.OnMouseMove(e);

			// Exit if the mouse is down over a SB
			if( m_nDSB!=SB.None ) return;

			bool bRedraw = false;

			// Is the mouse over a scroll button
			SB nSB = SBHitTest(e.X, e.Y);
			if( nSB!=m_nSB ) 
			{
				m_nSB = nSB;
				if( nSB!=SB.None ) UpdateHot();
				bRedraw = true;
			}

			// Update the hot tab, if not SB is hot
			if( m_nSB==SB.None && UpdateHot(e.X, e.Y) ) bRedraw = true;
	
			// Cursor
			
			Cursor oCursor = (m_nSB!=SB.None ? m_SBCursor : (m_oHTab!=null ? m_HotCursor : m_Cursor));//(m_oHTab!=null ? ((draggedTab) ? Cursors.IBeam :   ) : m_Cursor)) 
			if( base.Cursor!=oCursor ) base.Cursor = oCursor;
			
			/*if (e.X < (this.Size.Width + 3) && e.X > (this.Size.Width  - 3))
			{
				base.Cursor = Cursors.VSplit;
				if (draggedTab)
				{
					this.SetClientSizeCore(e.X, this.Size.Height);
					bRedraw = true;
				}
			}*/

			/*if (e.X < 3)// && e.X > (this.Location.X  - 3))
			{
				base.Cursor = Cursors.VSplit;
				if (draggedTab)
				{
					//this.Location = new Point(e.X, this.Location.Y);
					this.SetBounds(this.Location.X this.Size.Width - e.X , this.Size.Height);
					bRedraw = true;
				}
			}*/
			// Redraw the control if needed
			if( bRedraw ) Invalidate();
		}


		/*************************************************************************************
		OnMouseDown Override
		*************************************************************************************/
		protected override void OnMouseDown(MouseEventArgs e)
		{			
			// Is the mouse over a SB
			
			if( m_nSB!=SB.None ) 
			{
				if( m_bFirstVisible && (m_nSB==SB.First || m_nSB==SB.Prev) ) return;
				if( m_bLastVisible && (m_nSB==SB.Next || m_nSB==SB.Last) ) return;
				m_nDSB = m_nSB;
				Invalidate();
				return;
			}

			// Get the item under the mouse
			Tab oATab = HitTest(e.X, e.Y);
			
			/*if (oATab != null)
			{
				m_oDTab = oATab;
				this.Parent.DoDragDrop(m_oDTab, DragDropEffects.All);
				//
			}
			draggedTab = true;*/
			bool bRedraw = false;

			// Is the mouse over a new item?
			if( m_oATab!=oATab && oATab!=null )
			{
				if( OnTabChange(oATab, m_oATab)==false )
				{
					// Remove the active state from the current active tab
					if( m_oATab!=null ) m_oATab.DisableState(TabState.Active);

					// Change the active tab
					oATab.EnableState(TabState.Active);
					m_oATab = oATab;

					bRedraw = true;				
				}
				
			}

			// If the active tab is not fully visible make it fully visible
			if( oATab!=null && !m_bLastFullyVisible && m_nFirstIndex!=m_nLastIndex && oATab.Index==m_nLastIndex ) 
			{
				EnsureLast(m_nLastIndex);
				bRedraw = true;
			}

			if( bRedraw ) Invalidate();
			base.OnMouseDown(e);
			/*if(this.Parent != null)
				this.Parent.MouseDown += new MouseEventHandler(mouseParent_Down);*/
		}
/*--------------------------------------------------------------------------------------
		protected override void OnDragOver(DragEventArgs drgevent)
		{
			//drgevent.AllowedEffect = DragDropEffects.Move;
			base.OnDragOver (drgevent);
		}

//**************************************************************************************/

		/*************************************************************************************
		OnMouseUp Override
		*************************************************************************************/
		protected override void OnMouseUp(MouseEventArgs e)
		{
			draggedTab = false;
			if( m_nDSB!=SB.None )
			{
				// Was a SB clicked?
				SB nSB = SBHitTest(e.X, e.Y);
				if( nSB==m_nDSB ) ClickScrollButton(m_nDSB);

				// Update the control
				m_nDSB = SB.None;
				Invalidate();
			}
			//draggedTab = false;
			/*if(leaveTab && draggedTab)
			{
				
				leaveTab = false;
				CreateForm();
				Invalidate();
			}*/
					
			if (m_oDTab != null) m_oDTab = null;
			//MessageBox.Show("Mouse Up!!!");
		}

		
		/*************************************************************************************
		OnMouseLeave Override
		*************************************************************************************/
		protected override void OnMouseLeave(EventArgs e)
		{
	
			// Release the hot tab
			if( UpdateHot() ) Invalidate();

			// Release the hot SB
			if( m_nSB!=SB.None ) 
			{
				m_nSB = SB.None;
				Invalidate();
			}
			base.OnMouseLeave(e);
			//leaveTab = true;
		}

		/*************************************************************************************
		OnMouseHover Override
		*************************************************************************************/
		/*protected override void OnMouseHover(EventArgs e)
		{
			base.OnMouseHover(e);
			//leaveTab = false;
			//MessageBox.Show("DragDrop");
			
		}*/
		
		/*protected void CreateForm()
		{
			if (m_oDTab != null)
			{
				Form newForm = new Form();
				newForm.Text = m_oDTab.Text;
				newForm.FormBorderStyle = FormBorderStyle.SizableToolWindow;
				newForm.Size = new Size(this.Size.Width, 400);
				//Container formContainer = new Container();
				MyTabControl.TabCtl tabCtl1 = new MyTabControl.TabCtl();
				tabCtl1.BackColor = System.Drawing.Color.FromArgb(((System.Byte)(192)), ((System.Byte)(192)), ((System.Byte)(255)));
				tabCtl1.BottomMargin = 0;
				tabCtl1.Direction = MyTabControl.TabDirection.Bottom;
				tabCtl1.HorLineColor = System.Drawing.Color.Black;
				tabCtl1.HotTabCursor = System.Windows.Forms.Cursors.Hand;
				tabCtl1.LastVisibleTab = 0;
				tabCtl1.Dock = DockStyle.Bottom;
					//Location = new System.Drawing.Point(0, 0);
				tabCtl1.Name = "tabCtl1";
				tabCtl1.SBBackColor = System.Drawing.SystemColors.Control;
				tabCtl1.SBBorderStyle = System.Windows.Forms.Border3DStyle.RaisedInner;
				tabCtl1.SBDisableBackColor = System.Drawing.SystemColors.Control;
				tabCtl1.SBDisableForeColor = System.Drawing.Color.DarkGray;
				tabCtl1.SBForeColor = System.Drawing.Color.Black;
				tabCtl1.SBHotBackColor = System.Drawing.SystemColors.Control;
				tabCtl1.SBHotForeColor = System.Drawing.Color.Blue;
				tabCtl1.SBWidth = ((short)(17));
				tabCtl1.SeparatorColor = System.Drawing.Color.DarkGray;
				tabCtl1.Size = new System.Drawing.Size(200, 24);
				tabCtl1.TabIndex = 1;
				tabCtl1.TabSpacing = -3;
				tabCtl1.TopMargin = 2;
				//this.ImageList.Images.
				
				//this.Tabs.Reindex();
				
				//this.Refresh();
				m_oDTab.m_oParent.Tabs.RemoveAt(0);
				m_oDTab.m_oParent = tabCtl1;
				
				tabCtl1.Tabs.All = new Tab[1] {m_oDTab};
				newForm.Controls.Add(tabCtl1);
				newForm.Show();
			}
		}*/

		/*************************************************************************************
		Updates the Hot tab
		*************************************************************************************/
		protected bool UpdateHot(int X, int Y)
		{
			bool bRedraw = false;
			bool bNonHot = true;

			// Get the hot tab
			Tab oTab = HitTest(X, Y);	
			if( oTab!=null )
			{
				bNonHot = false;
				if( m_oHTab!=oTab )
				{
					// Remove the Hot state from the previous hot tab and set it for the new
					if( m_oHTab!=null ) m_oHTab.DisableState(TabState.Hot); 
					oTab.EnableState(TabState.Hot);
					m_oHTab = oTab;
					bRedraw = true;
				}
			}

			// Check if the previous hot tab is still hot
			if( bNonHot && m_oHTab!=null ) 
			{
				m_oHTab.DisableState(TabState.Hot);
				m_oHTab = null;
				bRedraw = true;
			}

			return bRedraw;
		}
		protected bool UpdateHot()
		{
			if( m_oHTab!=null ) 
			{
				m_oHTab.DisableState(TabState.Hot);
				m_oHTab = null;
				return true;
			}

			return false;
		}


		/*************************************************************************************
		Draws the control background
		*************************************************************************************/
		protected virtual void OnDrawBackground(System.Drawing.Graphics g)
		{
			g.FillRectangle(new SolidBrush(BackColor), ClientRectangle);
		}

		/*************************************************************************************
		Draws the control border
		*************************************************************************************/
		protected virtual void OnDrawBorder(System.Drawing.Graphics g)
		{
			ControlPaint.DrawBorder3D(g, ClientRectangle, m_Border);
		}

		/*************************************************************************************
		Draws the horizontal line
		*************************************************************************************/
		protected virtual void OnDrawHorizontalLine(System.Drawing.Graphics g)
		{
			// Line brush
			SolidBrush oBrush = new SolidBrush(m_HorLineColor);

			// Draw the line
			int Y = (m_nDirection==TabDirection.Top ? m_TabsArea.Y-m_nHorLineWidth:m_TabsArea.Bottom);
			g.FillRectangle(oBrush, m_TabsArea.X, Y, m_TabsArea.Width+1, m_nHorLineWidth);

			// Draw the active tab line with the same color as the active tab background color
			if( m_oATab!=null && IsTabVisible(m_oATab) ) 
			{
				Point[] aCoordinates = m_oATab.Coordinates;
				if( m_nDirection==TabDirection.Top )
					g.FillRectangle(new SolidBrush(m_oATab.ActiveBackColor), aCoordinates[0].X+1, Y, aCoordinates[1].X-aCoordinates[0].X-2, m_nHorLineWidth);
				else
					g.FillRectangle(new SolidBrush(m_oATab.ActiveBackColor), aCoordinates[3].X+1, Y, aCoordinates[2].X-aCoordinates[3].X-2, m_nHorLineWidth);
			}

			// If the SB area aligns with the tab area draw the horizontal line over the SB area
			if( (m_nDirection==TabDirection.Top && m_nTopMargin==0) || (m_nDirection==TabDirection.Bottom && m_nBottomMargin==0) ) 
				g.FillRectangle(oBrush, m_SBArea.X, Y, m_SBArea.Width+1, m_nHorLineWidth);

			// Clean up
			oBrush.Dispose();
		}

		/*************************************************************************************
		Draws tab separators
		*************************************************************************************/
		protected virtual void OnDrawSeparator(System.Drawing.Graphics g, Tab oTab)
		{
			int nX = Math.Min(oTab.Coordinates[1].X, oTab.Coordinates[2].X) - 1;		
			g.DrawLine(m_oSepPen, nX, oTab.Y+2, nX, oTab.Y+oTab.Height-4);
		}

		/*************************************************************************************
		Draws a scroll button
		*************************************************************************************/
		protected virtual void OnDrawSB(System.Drawing.Graphics g, SB nSB, int X, int Y, int nWidth, int nHeight)
		{
			// Validate the specified scroll button
			if( nSB!=SB.First && nSB!=SB.Prev && nSB!=SB.Next && nSB!=SB.Last ) return;

			// Colors and brushes
			Color nFore = (m_nSB==nSB ? m_SBHotFore:m_SBFore);
			Color nBack = (m_nSB==nSB ? m_SBHotBack:m_SBBack);
			if( m_bFirstVisible && (nSB==SB.First || nSB==SB.Prev) ) 
			{
				nFore = m_SBDisableFore;
				nBack = m_SBDisableBack;
			}
			if( m_bLastVisible && (nSB==SB.Next || nSB==SB.Last) ) 
			{
				nFore = m_SBDisableFore;
				nBack = m_SBDisableBack;
			}
			Brush oForeBrush = new SolidBrush(nFore);
			Brush oBackBrush = new SolidBrush(nBack);		

			// Background
			g.FillRectangle(oBackBrush, X, Y, nWidth, nHeight);

			int nOffset = (m_nDSB==nSB ? 1:0);

			// Triangle coodinates
			Point[] aCoor = null;
			if( nSB==SB.First || nSB==SB.Prev ) 
			{
				aCoor = new Point[] {	new Point( X+nWidth/2-2+nOffset, Y+nHeight/2+nOffset ),
										new Point( X+nWidth/2+2+nOffset, Y+nHeight/2-4+nOffset ),
										new Point( X+nWidth/2+2+nOffset, Y+nHeight/2+4+nOffset ) 
									};
			}
			else
			{
				aCoor = new Point[] {	new Point( X+nWidth/2-2+nOffset, Y+nHeight/2-4+nOffset ),
										new Point( X+nWidth/2+2+nOffset, Y+nHeight/2+nOffset ),
										new Point( X+nWidth/2-2+nOffset, Y+nHeight/2+4+nOffset ) 
									};		
			}

			// Draw the triangle
			g.FillPolygon(oForeBrush, aCoor);

			// The first and last button line before/after the triange
			if( nSB==SB.First )
				g.FillRectangle(oForeBrush, X+nWidth/2-4+nOffset, Y+nHeight/2-3+nOffset, 1, 7);
			else if( nSB==SB.Last )
				g.FillRectangle(oForeBrush, X+nWidth/2+3+nOffset, Y+nHeight/2-3+nOffset, 1, 7);

			// Border		
			Border3DStyle nBorder = (m_nDSB==nSB ? m_SBABorder: (m_nSB==nSB ? m_SBHBorder:m_SBBorder));
			ControlPaint.DrawBorder3D(g, X, Y, nWidth, nHeight, nBorder);

			// Clean up
			oForeBrush.Dispose();
			oBackBrush.Dispose();
		}
	

		/*************************************************************************************
		Draws all visible scroll buttons
		*************************************************************************************/
		protected virtual void OnDraAllSB(System.Drawing.Graphics g)
		{
			// Exit if no SB is visible
			if( m_nSBVisible==SB.None ) return;

			// Coordinates
			int nX = m_SBArea.X,
				nY = m_SBArea.Y,
				nHeight = m_SBArea.Height;

			// SB First
			if( (m_nSBVisible & SB.First)==SB.First ) 
			{
				OnDrawSB(g, SB.First, nX, nY, m_nSBWidth, nHeight);
				nX += m_nSBWidth;
			}
			// SB Prev
			if( (m_nSBVisible & SB.Prev)==SB.Prev ) 
			{
				OnDrawSB(g, SB.Prev, nX, nY, m_nSBWidth, nHeight);
				nX += m_nSBWidth;
			}
			// SB Next
			if( (m_nSBVisible & SB.Next)==SB.Next ) 
			{
				OnDrawSB(g, SB.Next, nX, nY, m_nSBWidth, nHeight);
				nX += m_nSBWidth;
			}
			// SB Last
			if( (m_nSBVisible & SB.Last)==SB.Last ) 
			{
				OnDrawSB(g, SB.Last, nX, nY, m_nSBWidth, nHeight);
				nX += m_nSBWidth;
			}
		}


		/*************************************************************************************
		Scroll button was clicked
		*************************************************************************************/
		protected void ClickScrollButton(SB nSB)
		{
			// Raise the event, exit if canceled
			if( OnSBClick(nSB) ) return;

			// What button?
			switch( nSB )
			{
				case SB.First: 
					EnsureFirst(0);
					break;
				case SB.Prev: 
					EnsureFirst(m_nFirstIndex-1);
					break;
				case SB.Next: 
					EnsureFirst(m_nFirstIndex+1);
					break;
				case SB.Last: 
					EnsureLast(m_oTabs.m_List.Count-1);
					break;
			}
		}
	

		/*************************************************************************************
		Returns the border size
		*************************************************************************************/
		protected virtual short BorderSize 
		{ 
			get 
			{ 
				switch( m_Border ) 
				{
					case Border3DStyle.Adjust:
						return 0;
					case Border3DStyle.Bump:
					case Border3DStyle.Etched:
					case Border3DStyle.Raised:
					case Border3DStyle.Sunken:
						return (short)SystemInformation.Border3DSize.Width;
					default:
						return (short)SystemInformation.BorderSize.Width;
				}
			} 
		}


		/*************************************************************************************
		Updates internal areas; tabs and scroll buttons
		*************************************************************************************/
		protected virtual void UpdateAreas()
		{
			short nBorderSize = BorderSize;
		
			// Number of visible scroll buttons
			int nVCount = 0;
			if( (m_nSBVisible & SB.First)==SB.First  ) nVCount++;
			if( (m_nSBVisible & SB.Prev)==SB.Prev  ) nVCount++;
			if( (m_nSBVisible & SB.Next)==SB.Next  ) nVCount++;
			if( (m_nSBVisible & SB.Last)==SB.Last  ) nVCount++;

			// Scroll buttons area
			m_SBArea.Y = nBorderSize + (m_nTopMargin==0 ? (m_bHorLine && m_nDirection==TabDirection.Top ? (int)m_nHorLineWidth:0):0);
			m_SBArea.Height = ClientRectangle.Height - nBorderSize - m_SBArea.Y - (m_nBottomMargin==0 ? (m_bHorLine && m_nDirection==TabDirection.Bottom ? (int)m_nHorLineWidth:0):0);
			m_SBArea.Width = nVCount * m_nSBWidth;
			m_SBArea.X = (m_nSBPosition==SBPos.Left ? nBorderSize : ClientRectangle.Right-m_SBArea.Width-nBorderSize);

			// Tabs area
			m_TabsArea.X = nBorderSize + m_nLeftMargin + (m_nSBPosition==SBPos.Left ? m_SBArea.Width:0);
			m_TabsArea.Y = nBorderSize + m_nTopMargin + (m_bHorLine && m_nDirection==TabDirection.Top ? (int)m_nHorLineWidth:0);
			m_TabsArea.Width = ClientRectangle.Width - (nBorderSize*2) - m_nRightMargin - m_nLeftMargin - m_SBArea.Width;
			m_TabsArea.Height = ClientRectangle.Height - nBorderSize - m_nBottomMargin - m_TabsArea.Y - (m_bHorLine && m_nDirection==TabDirection.Bottom ? (int)m_nHorLineWidth:0);
		}


		/*************************************************************************************
		Returns the previous or next visible tab from the specified index
		*************************************************************************************/
		protected int GetNextVisible(int nStartIndex)
		{
			for( int nLoop=nStartIndex; nLoop < m_oTabs.Count; nLoop++ ) 
			{
				if( ((Tab)m_oTabs.m_List[nLoop]).Visible ) return nLoop;
			}
			return -1;
		}
		protected int GetPrevVisible(int nStartIndex)
		{
			for( int nLoop=nStartIndex; nLoop >= 0 ; nLoop-- ) 
			{
				if( ((Tab)m_oTabs.m_List[nLoop]).Visible ) return nLoop;
			}
			return -1;
		}


		/*************************************************************************************
		Returns the number of tabs with the visible property of true
		*************************************************************************************/
		protected int GetVisibleCount()
		{
			int nReturn = 0;
			foreach( Tab oTab in m_oTabs ) if( oTab.Visible ) nReturn++;
			return nReturn;
		}


		/*************************************************************************************
		Ensures that the specified tab is the first/last visible. Returns the first/last
		tab index; -1 if a invalid tab was specified

		NOTE: There is no guarantee that a tab will be first/last, the only guarantee is 
			  that it will be visible. This only happens if the last/first tab is visible.
		*************************************************************************************/
		public bool EnsureFirst(int nIndex)
		{
			// Validate the specified index
			if( nIndex<0 || nIndex>=m_oTabs.m_List.Count ) 
			{
				if( m_oTabs.m_List.Count==0 ) 
				{
					m_bFirstVisible = true;
					m_bLastVisible= true;
				}
				return false;
			}			

			Tab oTab = null;
			int nFV = GetNextVisible(0);
			int nAreaWidth = m_TabsArea.Width-m_nInitTabSpace;

			// Get the start index
			nIndex = GetNextVisible(nIndex);
			if( nIndex==-1 ) return false;	// No visible tabs from the specified index

			m_nFirstIndex = nIndex;
	
			// Loop until the last tab is found
			while( true )
			{
				m_nLastIndex = nIndex;
				oTab = (Tab)m_oTabs.m_List[nIndex];
				nAreaWidth -= (oTab.Width+m_nTabSpacing);
				if( nAreaWidth<=0 ) break;
				nIndex++;
				nIndex = GetNextVisible(nIndex);
				if( nIndex==-1 && nFV==m_nFirstIndex ) break;
				if( nIndex==-1 ) return EnsureLast(m_oTabs.m_List.Count-1);
			}

			// Is the last visible tab fully visible
			m_bLastFullyVisible = (nAreaWidth>=0);

			// Update the visible tabs
			UpdateVisibleTabs();

			return true;
		}
		public bool EnsureLast(int nIndex)
		{
			// Validate the specified index
			if( nIndex<0 || nIndex>=m_oTabs.m_List.Count ) 
			{
				if( m_oTabs.m_List.Count==0 ) 
				{
					m_bFirstVisible = true;
					m_bLastVisible= true;
				}
				return false;
			}			

			// Get the starting index
			nIndex = GetPrevVisible(nIndex);
			if( nIndex==-1 ) return false;	// No visible tabs from the specified index

			m_bLastFullyVisible = true;
			m_nLastIndex = nIndex;

			// Loop until the first tab is found
			Tab oTab;
			int nAreaWidth = m_TabsArea.Width-m_nInitTabSpace;
			while( true )
			{
				oTab = (Tab)m_oTabs.m_List[nIndex];
				nAreaWidth -= (oTab.Width+m_nTabSpacing);

				// No more space available; the current tab does not fit
				if( nAreaWidth<0 ) 
				{
					m_nFirstIndex = GetNextVisible(nIndex+1);
					m_nLastIndex = GetNextVisible(m_nLastIndex+1);
					if( m_nFirstIndex==-1 )	m_nFirstIndex = GetPrevVisible(m_oTabs.Count-1);
					if( m_nLastIndex==-1 )	m_nLastIndex = GetPrevVisible(m_oTabs.Count-1);
					else					m_bLastFullyVisible = false;
					break;
				}
				// No more space available; the current tab fits perfectly
				if( nAreaWidth==0 ) 
				{
					m_nFirstIndex = nIndex;
					break;
				}

				nIndex--;
				nIndex = GetPrevVisible(nIndex);
				if( nIndex==-1 ) return EnsureFirst(0);
			}

			// Update the visible tabs
			UpdateVisibleTabs();

			return true;
		}


		/*************************************************************************************
		Updates the visible tabs
		*************************************************************************************/
		protected void UpdateVisibleTabs()
		{
			int nLeft = m_TabsArea.Left+m_nInitTabSpace;
			int nTop = m_TabsArea.Top;
			int nHeight = m_TabsArea.Height;
			Tab oTab = null;

			m_nVisibleCount = 0;

			// Set the current visible tabs as invisible
			for( int nLoop=0; nLoop<m_aVisible.Count; nLoop++ ) ((Tab)m_aVisible[nLoop]).EnableState(TabState.Invisible);
			m_aVisible.Clear();

			// Loop all the new visible tabs
			for( int nLoop=m_nFirstIndex; nLoop<=m_nLastIndex; nLoop++ )
			{
				oTab = (Tab)m_oTabs.m_List[nLoop];
				if( oTab.Visible ) 
				{
					// Update the tab
					oTab.Update(nLeft, nTop, nHeight, -1, TabState.Normal);
					if( m_oATab==oTab ) oTab.EnableState(TabState.Active);
					m_aVisible.Add(oTab);
				
					m_nVisibleCount++;

					// Increment the left position for the next tab
					nLeft += (oTab.Width+m_nTabSpacing);
				}
			}
		
			// Is the first/last tab visible
			m_bFirstVisible = GetNextVisible(0)==m_nFirstIndex;
			m_bLastVisible = GetPrevVisible(m_oTabs.m_List.Count-1)==m_nLastIndex && m_bLastFullyVisible;

			if( m_bReverseZOrder ) m_aVisible.Reverse();
		}


		/*************************************************************************************
		Called by a tab when a property changes
		*************************************************************************************/
		internal void TabChanged(Tab oTab, string sProperty, bool bFullUpdate)
		{
			// Update the control if auto update is enabled
			if( m_bAutoUpdate )
			{
				// Refresh the control if the tab is visible
				if( m_aVisible.IndexOf(oTab)!=-1 || sProperty=="Visible" )
				{
					if( bFullUpdate ) 	Refresh();
					else				Invalidate();
				}
			}

		}


		/*************************************************************************************
		Called by the tabs collection to update the control
		*************************************************************************************/
		internal void TabsChanged()
		{
			if( m_bAutoUpdate ) Refresh();
		}


		/*************************************************************************************
		Refreshes the control
		*************************************************************************************/
		public override void Refresh()
		{
			// Validate the first index
			if( m_nFirstIndex<0 || m_nFirstIndex>=m_oTabs.Count ) m_nFirstIndex = 0;

			// Update the areas
			UpdateAreas();

			// Refresh the visible tabs
			EnsureFirst(m_nFirstIndex);

			base.Refresh();
		}

		/*************************************************************************************
		Returns the tab at the specified position
		*************************************************************************************/
		public Tab HitTest(int X, int Y)
		{
			// Check the active tab first
			if( m_oATab!=null && m_oATab.HitTest(X, Y) ) return m_oATab;

			// Check the visible tabs
			Tab oTab = null;
			for( int nLoop=m_aVisible.Count-1; nLoop>=0; nLoop--) 
			{
				oTab = (Tab)m_aVisible[nLoop];
				if( oTab.HitTest(X, Y) ) return oTab;
			}
			return null;
		}
		public Tab HitTest(Point pos)
		{
			return HitTest(pos.X, pos.Y);
		}


		/*************************************************************************************
		Returns the scroll button at the specified position
		*************************************************************************************/
		public SB SBHitTest(int X, int Y)
		{
			// Is the coordinate in the SB area
			if( m_SBArea.Contains(X, Y) )
			{
				int nX = m_SBArea.X;

				// SB First
				if( (m_nSBVisible & SB.First)==SB.First && X<=(nX+m_nSBWidth) ) 
					return SB.First;
				nX += m_nSBWidth;
				// SB Prev
				if( (m_nSBVisible & SB.Prev)==SB.Prev && X<=(nX+m_nSBWidth) ) 
					return SB.Prev;
				nX += m_nSBWidth;
				// SB Next
				if( (m_nSBVisible & SB.Next)==SB.Next && X<=(nX+m_nSBWidth) ) 
					return SB.Next;
				nX += m_nSBWidth;
				// SB Last
				if( (m_nSBVisible & SB.Last)==SB.Last && X<=(nX+m_nSBWidth) ) 
					return SB.Last;
			}

			return SB.None;
		}
		public SB SBHitTest(Point pos)
		{
			return SBHitTest(pos.X, pos.Y);
		}

	
		/*************************************************************************************
		Returns true if a tab is visible
		*************************************************************************************/
		public bool IsTabVisible(int nIndex)
		{
			if( nIndex<m_nFirstIndex || nIndex>m_nLastIndex ) return false;
			return m_oTabs[nIndex].Visible;
		}
		public bool IsTabVisible(Tab oTab)
		{
			return m_aVisible.IndexOf(oTab)!=-1;
		}

		/*************************************************************************************
		Returns the tabs collection
		*************************************************************************************/
		public bool ShouldPersistTabs() { return true; }
		[Category("Appearance"), Editor(typeof(TabsCollectionEditor), typeof(System.Drawing.Design.UITypeEditor)), ]
		public MyTabControl.Tabs Tabs 
		{
			get { return m_oTabs; }
		}
		
		/*************************************************************************************
		Get/Set the first/last visible tabs
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(0), ]
		public int FirstVisibleTab 
		{
			get { return m_nFirstIndex; }
			set 
			{ 
				EnsureFirst(value); 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("FirstVisibleTab");
			}
		}
		[Browsable(false), ]//Persistable(PersistableSupport.None)]
		public int LastVisibleTab 
		{
			get { return m_nLastIndex; }
			set 
			{ 
				EnsureLast(value); 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("LastVisibleTab");
			}
		}

		/*************************************************************************************
		Returns true if the last visible tab is fully visible
		*************************************************************************************/
		[Browsable(false)]
		public bool IsLastTabFullyVisible 
		{
			get { return m_bLastFullyVisible; }
		}

		/*************************************************************************************
		Returns true if the last visible tab is fully visible
		*************************************************************************************/
		[Browsable(false)]
		public int VisibleTabsCount 
		{
			get { return m_nVisibleCount; }
		}

		/*************************************************************************************
		Returns the visible tabs array
		*************************************************************************************/
		[Browsable(false)]
		public ArrayList VisibleTabs 
		{
			get { return m_aVisible; }
		}

		/*************************************************************************************
		Returns the areas
		*************************************************************************************/
		[Browsable(false)]
		public Rectangle TabsArea 
		{
			get { return m_TabsArea; }
		}
		[Browsable(false)]
		public Rectangle SBArea 
		{
			get { return m_SBArea; }
		}

		/*************************************************************************************
		Get/Set the AutoUpdate property
		*************************************************************************************/
		[Category("Behavior"), Bindable(false), DefaultValue(true), ]
		public bool AutoUpdate 
		{
			get { return m_bAutoUpdate; }
			set 
			{ 
				if( m_bAutoUpdate==value ) return;
				m_bAutoUpdate = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("AutoUpdate");
			}
		}

		/*************************************************************************************
		Get/Set back ground color property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), ]
		public override Color BackColor
		{
			get { return base.BackColor; }
			set 
			{ 
				base.BackColor = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("BackColor");
			}
		}
		public bool ShouldPersistBackColor() { return !(base.BackColor==Color.FromArgb(247, 243, 233)); }
		public override void ResetBackColor() { base.BackColor=Color.FromArgb(247, 243, 233); }

		/*************************************************************************************
		Get/Set the control margins
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), ]
		public int TopMargin 
		{
			get { return m_nTopMargin; }
			set 
			{ 
				if( m_nTopMargin==value ) return;
				m_nTopMargin = Int16.Parse(value.ToString());
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("TopMargin");
			}
		}
		public bool ShouldPersistTopMargin() { return m_nTopMargin!=(m_nDirection==TabDirection.Top ? 0:2); }
		public void ResetTopMargin() { TopMargin = (m_nDirection==TabDirection.Top ? 0:2); }
		[Category("Appearance"), Bindable(false), DefaultValue(0), ]
		public int LeftMargin 
		{
			get { return m_nLeftMargin; }
			set 
			{ 
				if( m_nLeftMargin==value ) return;
				m_nLeftMargin = Int16.Parse(value.ToString());//.ToInt16(); 
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("LeftMargin");
			}
		}
		[Category("Appearance"), Bindable(false), DefaultValue(0), ]
		public int RightMargin 
		{
			get { return m_nRightMargin; }
			set 
			{ 
				if( m_nRightMargin==value ) return;
				m_nRightMargin = Int16.Parse(value.ToString());//.ToInt16(); 
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("RightMargin");
			}
		}
		[Category("Appearance"), Bindable(false), ]
		public int BottomMargin 
		{
			get { return m_nBottomMargin; }
			set 
			{ 
				if( m_nBottomMargin==value ) return;
				m_nBottomMargin = Int16.Parse(value.ToString());//.ToInt16(); 
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("BottomMargin");
			}
		}
		public bool ShouldPersistBottomMargin() { return m_nBottomMargin!=(m_nDirection==TabDirection.Top ? 2:0); }
		public void ResetBottomMargin() { BottomMargin = (m_nDirection==TabDirection.Top ? 2:0); }

		/*************************************************************************************
		Get/Set Horizontal line properties
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(true), ]
		public bool DrawHorLine 
		{
			get { return m_bHorLine; }
			set 
			{ 
				if( m_bHorLine==value ) return;
				m_bHorLine = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("DrawHorLine");
			}
		}
		[Category("Appearance"), Bindable(false), DefaultValue(1), ]
		public int HorLineWidth 
		{
			get { return m_nHorLineWidth; }
			set 
			{ 
				if( m_nHorLineWidth==value ) return;
				m_nHorLineWidth = Int16.Parse(value.ToString());//.ToInt16(); 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("HorLineWidth");
			}
		}
		[Category("Appearance"), Bindable(false), ]
		public Color HorLineColor 
		{
			get { return m_HorLineColor; }
			set 
			{ 
				m_HorLineColor = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("HorLineColor");
			}
		}
		public bool ShouldPersistHorLineColor() { return !(m_HorLineColor==Color.Black); }
		public void ResetHorLineColor() { m_HorLineColor=Color.Black; }

		/*************************************************************************************
		Get/Set scroll buttons position property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(SBPos.Left), ]
		public SBPos SBPosition
		{
			get { return m_nSBPosition; }
			set 
			{ 
				if( m_nSBPosition==value ) return;
				m_nSBPosition = value;
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("SBPosition");
			}
		}

		/*************************************************************************************
		Get/Set scroll buttons width property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), ]
		public short SBWidth 
		{
			get { return m_nSBWidth; }
			set 
			{ 
				if( m_nSBWidth==value ) return;
				m_nSBWidth = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("SBWidth");
			}
		}
		public bool ShouldPersistSBWidth() { return !(m_nSBWidth==SystemInformation.VerticalScrollBarWidth); }
		public void ResetSBWidth() { m_nSBWidth=(short)SystemInformation.VerticalScrollBarWidth; }

		/*************************************************************************************
		Get/Set scroll buttons border style properties
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(Border3DStyle.Adjust), ]
		public Border3DStyle SBBorderStyle
		{
			get { return m_SBBorder; }
			set 
			{ 
				if( m_SBBorder==value ) return;
				m_SBBorder = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("SBBorderStyle");
			}
		}
		[Category("Appearance"), Bindable(false), DefaultValue(Border3DStyle.RaisedInner), ]
		public Border3DStyle SBHotBorderStyle
		{
			get { return m_SBHBorder; }
			set 
			{ 
				if( m_SBHBorder==value ) return;
				m_SBHBorder = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBHotBorderStyle");
			}
		}
		[Category("Appearance"), Bindable(false), DefaultValue(Border3DStyle.SunkenOuter), ]
		public Border3DStyle SBActiveBorderStyle
		{
			get { return m_SBABorder; }
			set 
			{ 
				if( m_SBABorder==value ) return;
				m_SBABorder = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBActiveBorderStyle");
			}
		}

		/*************************************************************************************
		Get/Set scroll buttons color properties
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), ]
		public Color SBBackColor
		{
			get { return m_SBBack; }
			set 
			{ 
				m_SBBack = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBBackColor");
			}
		}
		public bool ShouldPersistSBBackColor() { return !(m_SBBack==SystemColors.Control); }
		public void ResetSBBackColor() { m_SBBack=SystemColors.Control; }

		[Category("Appearance"), Bindable(false), ]
		public Color SBHotBackColor
		{
			get { return m_SBHotBack; }
			set 
			{ 
				m_SBHotBack = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBHotBackColor");
			}
		}
		public bool ShouldPersistSBHotBackColor() { return !(m_SBHotBack==SystemColors.Control); }
		public void ResetSBHotBackColor() { m_SBHotBack=SystemColors.Control; }

		[Category("Appearance"), Bindable(false), ]
		public Color SBForeColor
		{
			get { return m_SBFore; }
			set 
			{ 
				m_SBFore = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBForeColor");
			}
		}
		public bool ShouldPersistSBForeColor() { return !(m_SBFore==Color.Black); }
		public void ResetSBForeColor() { m_SBFore=Color.Black; }

		[Category("Appearance"), Bindable(false), ]
		public Color SBHotForeColor
		{
			get { return m_SBHotFore; }
			set 
			{ 
				m_SBHotFore = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBHotForeColor");
			}
		}
		public bool ShouldPersistSBHotForeColor() { return !(m_SBHotFore==Color.Blue); }
		public void ResetSBHotForeColor() { m_SBHotFore=Color.Blue; }

		[Category("Appearance"), Bindable(false), ]
		public Color SBDisableForeColor
		{
			get { return m_SBDisableFore; }
			set 
			{ 
				m_SBDisableFore = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBDisableForeColor");
			}
		}
		public bool ShouldPersistSBDisableForeColor() { return !(m_SBDisableFore==Color.DarkGray); }
		public void ResetSBDisableForeColor() { m_SBDisableFore=Color.DarkGray; }

		[Category("Appearance"), Bindable(false), ]
		public Color SBDisableBackColor
		{
			get { return m_SBDisableBack; }
			set 
			{ 
				m_SBDisableBack = value; 
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SBDisableBackColor");
			}
		}
		public bool ShouldPersistSBDisableBackColor() { return !(m_SBDisableBack==SystemColors.Control); }
		public void ResetSBDisableBackColor() { m_SBDisableBack=SystemColors.Control; }

		/*************************************************************************************
		Get/Set the visible SB buttons property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(SB.All), ]
		public SB SBVisible
		{
			get { return m_nSBVisible; }
			set 
			{ 
				m_nSBVisible = value;
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("SBVisible");
			}
		}

		/*************************************************************************************
		Get/Set separator color properties
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), ]
		public Color SeparatorColor
		{
			get { return m_SepColor; }
			set 
			{ 
				m_SepColor = value; 
				if( m_oSepPen!=null ) m_oSepPen.Dispose();
				m_oSepPen = new Pen(m_SepColor, 1);
				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("SeparatorColor");
			}
		}
		public bool ShouldPersistSeparatorColor() { return !(m_SepColor==Color.DarkGray); }
		public void ResetSeparatorColor() { m_SepColor=Color.DarkGray; }

		/*************************************************************************************
		Get/Set the BorderStyle property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(Border3DStyle.SunkenOuter), ]
		public Border3DStyle BorderStyle 
		{
			get { return m_Border; }
			set 
			{ 
				if( m_Border==value ) return;
				m_Border = value;
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("BorderStyle");
			}
		}

		/*************************************************************************************
		Get/Set the tab spacing property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(0), ]
		public int TabSpacing 
		{
			get { return m_nTabSpacing; }
			set 
			{
				if( m_nTabSpacing==value ) return;
				m_nTabSpacing = (short)value;
				if( m_bAutoUpdate ) 
				{
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("TabSpacing");
			}
		}

		/*************************************************************************************
		Get/Set the initial tab spacing property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(3), ]
		public int InitTabSpace 
		{
			get { return m_nInitTabSpace; }
			set 
			{
				if( m_nInitTabSpace==value ) return;
				m_nInitTabSpace = (short)value;
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("InitTabSpace");
			}
		}

		/*************************************************************************************
		Get/Set the reverse tab Zorder property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(true), ]
		public bool ReverseTabZOrder
		{
			get { return m_bReverseZOrder; }
			set 
			{ 
				m_bReverseZOrder = value;
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("ReverseTabZOrder");
			}
		}

		/*************************************************************************************
		Get/Set the hot tab cursor property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false),]
		public System.Windows.Forms.Cursor HotTabCursor 
		{
			get { return m_HotCursor;}
			set 
			{ 
				m_HotCursor=value;
				//RaisePropertyChangedEvent("HotTabCursor");
			}
		}
		public bool ShouldPersistHotTabCursor() { return m_HotCursor!=Cursors.Hand; }
		public void ResetHotTabCursor() { m_HotCursor=Cursors.Hand; }

		/*************************************************************************************
		Cursor property Override
		*************************************************************************************/
		[Category("Appearance"), Bindable(false),]
		public override System.Windows.Forms.Cursor Cursor 
		{ 
			get {return base.Cursor;}  
			set 
			{
				base.Cursor=value;
				m_Cursor=value;
				//RaisePropertyChangedEvent("Cursor");
			} 
		}

		/*************************************************************************************
		DoubleBuffer property
		*************************************************************************************/
		[Category("Behavior"), DefaultValue(true),]
		public bool DoubleBuffer 
		{
			get { return (m_oBuffer!=null);}
			set 
			{ 
				if( value )
				{
					// Init double buffer
					if( m_oBuffer==null ) m_oBuffer = new MyTabControl.Graphics.GraphicsBuffer();
				}
				else
				{
					// Remove double buffer
					if( m_oBuffer!=null ) 
					{
						m_oBuffer.Dispose();
						m_oBuffer = null;
					}
				}
				//RaisePropertyChangedEvent("DoubleBuffer");
			}
		}

		/*************************************************************************************
		Get/Set the draw separator property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(true),]
		public bool DrawSeparator 
		{
			get { return (m_oSepPen!=null);}
			set 
			{ 
				if( value )
				{
					if( m_oSepPen==null ) m_oSepPen = new Pen(m_SepColor, 1);
				}
				else
				{
					if( m_oSepPen!=null ) 
					{
						m_oSepPen.Dispose();
						m_oSepPen = null;
					}
				}
				//RaisePropertyChangedEvent("DrawSeparator");
			}
		}

		/*************************************************************************************
		Get/Set the active tab property
		*************************************************************************************/
		[Category("Appearance"), Bindable(false), DefaultValue(-1),]
		public int ActiveTab 
		{
			get { return (m_oATab==null ? -1:m_oATab.m_nIndex);}
			set 
			{ 
				// Get the new active tab; exit if a invalid tab was specified
				Tab oTab = null;
				try { oTab = m_oTabs[value]; } 
				catch{ oTab = null; }
				if( oTab==null && value!=-1 ) return;
			
				// Clear the current active tab
				if( m_oATab!=null && m_oATab.m_nIndex==value ) return;
				if( m_oATab!=null ) 
				{
					m_oATab.DisableState(TabState.Active);
					m_oATab = null;
				}

				// Set the new active tab
				m_oATab = oTab;
				if( m_oATab!=null ) m_oATab.EnableState(TabState.Active);

				if( m_bAutoUpdate ) Invalidate();
				//RaisePropertyChangedEvent("ActiveTab");
			}
		}


		/*************************************************************************************
		Get/Set the tab style property
		*************************************************************************************/
		[Category("Appearance"), Browsable(true), DefaultValue(TabStyle.VisualStudio)]
		public TabStyle TabStyle
		{
			get { return m_nTabStyle; }
			set 
			{ 
				// Set the style
				if( m_nTabStyle==value ) return;
				m_nTabStyle = value;

				// Avoid multiple updates
				bool bAutoUpdate = m_bAutoUpdate;
				m_bAutoUpdate = false;

				// Update style properties
				if( value==TabStyle.Excel || value==TabStyle.InterDev  ) 
				{
					TabSpacing = -7;
					DrawSeparator = false;
				} if( value==TabStyle.VisualStudio ) 
				  {
					  TabSpacing = 0;
					  DrawSeparator = true;
				  }

				// Update the tabs
				m_oTabs.UpdateStyle(value);

				// Update the control
				if( bAutoUpdate ) 
				{
					m_bAutoUpdate = true;
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("TabStyle");
			}
		}

		/*************************************************************************************
		Get/Set the tab direction property
		*************************************************************************************/
		[Category("Appearance"), Browsable(true), DefaultValue(TabDirection.Top)]
		public TabDirection Direction
		{
			get { return m_nDirection; }
			set 
			{ 
				if( m_nDirection==value ) return;
				m_nDirection = value;

				// Avoid multiple updates
				bool bAutoUpdate = m_bAutoUpdate;
				m_bAutoUpdate = false;

				// Flip the top and bottom margins
				int nValue = m_nTopMargin;
				TopMargin = m_nBottomMargin;
				BottomMargin = nValue;

				// Update the tabs
				m_oTabs.UpdateDirection();

				// Update the control
				if( bAutoUpdate ) 
				{
					m_bAutoUpdate = true;
					UpdateAreas();
					Refresh();
				}
				//RaisePropertyChangedEvent("Direction");
			}
		}

		/*************************************************************************************
		Get/Set image lists properties
		*************************************************************************************/
		[Category("Appearance"), Browsable(true), DefaultValue(null)]
		public System.Windows.Forms.ImageList ImageList
		{
			get { return m_ImageList; }
			set 
			{ 
				m_ImageList = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("ImageList");
			}
		}
		[Category("Appearance"), Browsable(true), DefaultValue(null)]
		public System.Windows.Forms.ImageList HotImageList
		{
			get { return m_HotImageList; }
			set 
			{ 
				m_HotImageList = value; 
				if( m_bAutoUpdate ) Refresh();
				//RaisePropertyChangedEvent("HotImageList");
			}
		}

	}
}